"""Protobuf serialization for YARA AST."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any
import warnings

from yaraast.errors import SerializationError
from yaraast.serialization.protobuf_conversion import (
    ast_to_protobuf,
    convert_expression_to_protobuf,
    convert_hex_token_to_protobuf,
    convert_rule_to_protobuf,
    convert_string_to_protobuf,
    protobuf_to_ast,
)
from yaraast.visitor.defaults import DefaultASTVisitor

from . import yara_ast_pb2

if TYPE_CHECKING:
    from yaraast.ast.base import YaraFile

# Suppress protobuf version warning before import
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message=".*Protobuf gencode version.*",
    module="google.protobuf.runtime_version",
)


class ProtobufSerializer(DefaultASTVisitor[Any]):
    """Protobuf serializer for YARA AST with efficient binary format."""

    def __init__(self, include_metadata: bool = True) -> None:
        super().__init__(default=None)
        self.include_metadata = include_metadata

    def serialize(self, ast: YaraFile, output_path: str | Path | None = None) -> bytes:
        """Serialize AST to Protobuf binary format."""
        pb_yara_file = self._ast_to_protobuf(ast)
        binary_data = pb_yara_file.SerializeToString()

        if output_path:
            with Path(output_path).open("wb") as f:
                f.write(binary_data)

        return binary_data

    def serialize_text(
        self,
        ast: YaraFile,
        output_path: str | Path | None = None,
    ) -> str:
        """Serialize AST to Protobuf text format (for debugging)."""
        pb_yara_file = self._ast_to_protobuf(ast)
        text_data = str(pb_yara_file)

        if output_path:
            with Path(output_path).open("w", encoding="utf-8") as f:
                f.write(text_data)

        return text_data

    def deserialize(
        self,
        binary_data: bytes | None = None,
        input_path: str | Path | None = None,
    ) -> YaraFile:
        """Deserialize Protobuf binary to AST."""
        if input_path:
            with Path(input_path).open("rb") as f:
                binary_data = f.read()

        if not binary_data:
            msg = "No binary data provided"
            raise SerializationError(msg)

        pb_yara_file = yara_ast_pb2.YaraFile()
        pb_yara_file.ParseFromString(binary_data)

        return self._protobuf_to_ast(pb_yara_file)

    def _ast_to_protobuf(self, ast: YaraFile) -> yara_ast_pb2.YaraFile:
        """Convert AST to Protobuf message."""
        return ast_to_protobuf(ast, include_metadata=self.include_metadata)

    def _convert_rule_to_protobuf(self, rule, pb_rule) -> None:
        """Convert rule AST to Protobuf."""
        convert_rule_to_protobuf(rule, pb_rule)

    def _convert_string_to_protobuf(self, string_def, pb_string) -> None:
        """Convert string definition to Protobuf."""
        convert_string_to_protobuf(string_def, pb_string)

    def _convert_hex_token_to_protobuf(self, token, pb_token) -> None:
        """Convert hex token to Protobuf."""
        convert_hex_token_to_protobuf(token, pb_token)

    def _convert_expression_to_protobuf(self, expr, pb_expr) -> None:
        """Convert expression to Protobuf."""
        convert_expression_to_protobuf(expr, pb_expr)

    def _protobuf_to_ast(self, pb_file: yara_ast_pb2.YaraFile) -> YaraFile:
        """Convert Protobuf message to AST."""
        return protobuf_to_ast(pb_file)

    def get_serialization_stats(self, ast: YaraFile) -> dict[str, Any]:
        """Get statistics about the serialization."""
        pb_file = self._ast_to_protobuf(ast)
        binary_size = len(pb_file.SerializeToString())
        text_size = len(str(pb_file))

        return {
            "binary_size_bytes": binary_size,
            "text_size_bytes": text_size,
            "compression_ratio": text_size / binary_size if binary_size > 0 else 0,
            "rules_count": len(ast.rules),
            "imports_count": len(ast.imports),
            "includes_count": len(ast.includes),
        }



"""Validation services for CLI (logic without IO)."""

from __future__ import annotations

from pathlib import Path

from yaraast.cli.utils import parse_yara_file
from yaraast.errors import ValidationError
from yaraast.libyara.cross_validator import CrossValidator
from yaraast.libyara.equivalence import EquivalenceTester
from yaraast.yarax.compatibility_checker import YaraXCompatibilityChecker
from yaraast.yarax.feature_flags import YaraXFeatures


def validate_rule_file(rule_file: str):
    ast = parse_yara_file(rule_file)
    return ast, len(ast.rules), len(ast.imports), sum(len(rule.strings) for rule in ast.rules)


def cross_validate_rules(rule_file: str, test_data: bytes, externals: dict[str, str]):
    content = Path(rule_file).read_text()
    from yaraast.parser.parser import Parser

    parser = Parser(content)
    ast = parser.parse()
    validator = CrossValidator()
    return validator.validate(ast, test_data, externals)


def roundtrip_test(rule_file: str, data: bytes | None):
    tester = EquivalenceTester()
    return tester.test_file_round_trip(rule_file, data)


def yarax_check(ast, strict: bool):
    features = YaraXFeatures.yarax_strict() if strict else YaraXFeatures.yarax_compatible()
    checker = YaraXCompatibilityChecker(features)
    return checker.check(ast)


def parse_externals(external: tuple[str, ...]) -> dict[str, str]:
    externals: dict[str, str] = {}
    for ext in external:
        if "=" not in ext:
            msg = f"Invalid external format: {ext}"
            raise ValidationError(msg)
        key, value = ext.split("=", 1)
        externals[key] = value
    return externals


def read_test_data(test_data_path: str | None) -> bytes | None:
    """Read test data if provided."""
    if not test_data_path:
        return None

    try:
        with Path(test_data_path).open("rb") as f:
            return f.read()
    except Exception as exc:
        msg = f"Error reading test data: {exc}"
        raise ValidationError(msg) from exc



"""Semantic validation CLI command."""

from __future__ import annotations

from pathlib import Path
import sys

import click

from yaraast.cli.semantic_reporting import (
    display_processing_error,
    display_summary,
    display_text_results,
    display_validation_start,
    emit_json_results,
    write_output_file,
)
from yaraast.cli.semantic_services import (
    _create_file_result,
    _create_validation_context,
    _exit_with_appropriate_code,
    _process_file,
)


@click.command()
@click.argument("files", nargs=-1, type=click.Path(exists=True, path_type=Path))
@click.option(
    "--output",
    "-o",
    type=click.Path(path_type=Path),
    help="Output file for validation results (JSON format)",
)
@click.option(
    "--format",
    "-f",
    type=click.Choice(["text", "json"]),
    default="text",
    help="Output format",
)
@click.option(
    "--warnings/--no-warnings",
    default=True,
    help="Include warnings in output",
)
@click.option(
    "--suggestions/--no-suggestions",
    default=True,
    help="Include suggestions in output",
)
@click.option(
    "--strict",
    is_flag=True,
    help="Treat warnings as errors (exit with non-zero code)",
)
@click.option(
    "--quiet",
    "-q",
    is_flag=True,
    help="Only show errors and warnings, not success messages",
)
def semantic(
    files: tuple[Path, ...],
    output: Path | None,
    format: str,
    warnings: bool,
    suggestions: bool,
    strict: bool,
    quiet: bool,
) -> None:
    """Perform semantic validation on YARA files."""
    if not files:
        click.echo("Error: No files provided", err=True)
        sys.exit(1)

    validation_context = _create_validation_context()
    all_results = []
    total_errors = 0
    total_warnings = 0

    for file_path in files:
        try:
            display_validation_start(file_path, quiet)
            result = _process_file(
                file_path, validation_context["parser"], validation_context["validator"]
            )

            file_result = _create_file_result(file_path, result)
            all_results.append(file_result)

            total_errors += len(result.errors)
            total_warnings += len(result.warnings)

            if format == "text":
                display_text_results(file_path, result, warnings, suggestions, quiet)
        except Exception as e:
            display_processing_error(file_path, e)
            continue

    if not quiet and format == "text":
        display_summary(len(files), total_errors, total_warnings)

    if output:
        write_output_file(output, all_results, format)
    elif format == "json":
        emit_json_results(all_results)

    _exit_with_appropriate_code(total_errors, total_warnings, strict)



"""Additional real coverage for expression type inference."""

from __future__ import annotations

from yaraast.ast.comments import Comment, CommentGroup
from yaraast.ast.conditions import AtExpression, ForOfExpression, InExpression, OfExpression
from yaraast.ast.expressions import (
    BinaryExpression,
    BooleanLiteral,
    Identifier,
    IntegerLiteral,
    RangeExpression,
    SetExpression,
    StringCount,
    StringIdentifier,
    StringLength,
    StringLiteral,
    StringOffset,
)
from yaraast.ast.extern import ExternImport, ExternNamespace, ExternRule, ExternRuleReference
from yaraast.ast.modules import DictionaryAccess
from yaraast.ast.operators import DefinedExpression, StringOperatorExpression
from yaraast.ast.pragmas import InRulePragma, Pragma, PragmaBlock, PragmaType
from yaraast.types._expr_inference import ExpressionTypeInference, _TypeBaseVisitor
from yaraast.types._registry import (
    BooleanType,
    IntegerType,
    StringType,
    TypeEnvironment,
    UnknownType,
)


def test_type_base_visitor_default_methods_return_unknown() -> None:
    visitor = _TypeBaseVisitor()

    nodes = [
        Comment("x"),
        CommentGroup([Comment("x")]),
        DefinedExpression(expression=IntegerLiteral(value=1)),
        StringOperatorExpression(
            left=StringLiteral(value="a"),
            operator="contains",
            right=StringLiteral(value="b"),
        ),
        ExternImport(module_path="ext"),
        ExternNamespace(name="ns"),
        ExternRule(name="r"),
        ExternRuleReference(rule_name="r"),
        InRulePragma(pragma=Pragma(pragma_type=PragmaType.PRAGMA, name="pragma")),
        Pragma(pragma_type=PragmaType.PRAGMA, name="pragma"),
        PragmaBlock(pragmas=[]),
    ]

    for node in nodes:
        out = visitor.visit(node)
        if isinstance(node, DefinedExpression | StringOperatorExpression):
            assert isinstance(out, BooleanType)
        else:
            assert isinstance(out, UnknownType)


def test_expr_inference_reports_undefined_string_variants() -> None:
    env = TypeEnvironment()
    inf = ExpressionTypeInference(env)

    assert isinstance(inf.infer(StringCount(string_id="missing")), UnknownType)
    assert isinstance(inf.infer(StringOffset(string_id="missing")), UnknownType)
    assert isinstance(inf.infer(StringLength(string_id="missing")), UnknownType)

    assert "Undefined string: $missing" in inf.errors[0]
    assert "Undefined string: $missing" in inf.errors[1]
    assert "Undefined string: $missing" in inf.errors[2]


def test_expr_inference_string_length_invalid_index_reports_error() -> None:
    env = TypeEnvironment()
    env.add_string("$a")
    inf = ExpressionTypeInference(env)

    out = inf.infer(StringLength(string_id="a", index=StringLiteral(value="bad")))
    assert isinstance(out, IntegerType)
    assert "String length index must be integer" in inf.errors[0]


def test_expr_inference_comparison_and_builtin_function_paths() -> None:
    env = TypeEnvironment()
    inf = ExpressionTypeInference(env)

    cmp_out = inf.infer(
        BinaryExpression(
            left=StringLiteral(value="x"),
            operator="==",
            right=IntegerLiteral(value=1),
        ),
    )
    assert isinstance(cmp_out, BooleanType)
    assert "Incompatible types for '=='" in inf.errors[0]

    right_bad = inf.infer(
        BinaryExpression(
            left=BooleanLiteral(value=True),
            operator="and",
            right=IntegerLiteral(value=1),
        ),
    )
    assert isinstance(right_bad, BooleanType)
    assert any("Right operand of 'and' must be boolean" in e for e in inf.errors)


def test_expr_inference_at_in_and_of_error_paths() -> None:
    env = TypeEnvironment()
    inf = ExpressionTypeInference(env)

    assert isinstance(
        inf.infer(AtExpression(string_id="$a", offset=StringLiteral(value="bad"))),
        BooleanType,
    )
    assert any("Offset in 'at' expression must be integer" in e for e in inf.errors)

    assert isinstance(
        inf.infer(InExpression(subject="$a", range=IntegerLiteral(value=1))),
        BooleanType,
    )
    assert any("'in' expression requires range" in e for e in inf.errors)

    assert isinstance(
        inf.infer(
            OfExpression(
                quantifier=BooleanLiteral(value=True),
                string_set=IntegerLiteral(value=1),
            ),
        ),
        BooleanType,
    )
    assert any("'of' quantifier must be string or integer" in e for e in inf.errors)
    assert any("'of' requires string set" in e for e in inf.errors)


def test_expr_inference_helper_and_branch_edges() -> None:
    env = TypeEnvironment()
    inf = ExpressionTypeInference(env)

    assert inf._normalize_string_id("a") == "$a"
    assert inf._normalize_string_id("$a") == "$a"

    env.modules.add("")
    assert inf._resolve_module_type("") is None

    env2 = TypeEnvironment()
    env2.add_module("ghost")
    assert ExpressionTypeInference(env2)._resolve_module_type("ghost") is None

    high_bad = ExpressionTypeInference(TypeEnvironment())
    out = high_bad.infer(
        InExpression(
            subject="$a",
            range=RangeExpression(
                low=IntegerLiteral(value=1),
                high=StringLiteral(value="x"),
            ),
        ),
    )
    assert isinstance(out, BooleanType)
    assert any("Range high bound must be integer" in e for e in high_bad.errors)

    dict_env = TypeEnvironment()
    from yaraast.types._registry import DictionaryType

    dict_env.define("d", DictionaryType(StringType(), IntegerType()))
    dict_out = ExpressionTypeInference(dict_env).infer(
        DictionaryAccess(object=Identifier(name="d"), key="plain-key"),
    )
    assert isinstance(dict_out, IntegerType)

    set_env = TypeEnvironment()
    set_env.add_string("$a")
    fo = ExpressionTypeInference(set_env).infer(
        ForOfExpression(
            quantifier=IntegerLiteral(value=1),
            string_set=SetExpression(elements=[StringIdentifier(name="$a")]),
            condition=None,
        ),
    )
    assert isinstance(fo, BooleanType)



"""Real tests for LSP document links (no mocks)."""

from __future__ import annotations

from textwrap import dedent

from yaraast.lsp.document_links import DocumentLinksProvider
from yaraast.lsp.document_types import path_to_uri
from yaraast.lsp.runtime import DocumentContext


def test_document_links_import_and_include(tmp_path) -> None:
    include_path = tmp_path / "inc.yar"
    include_path.write_text("rule inc { condition: true }")

    text = dedent(
        f"""
        import "pe"
        include "{include_path.name}"

        rule main {{
            condition:
                pe.is_pe
        }}
        """,
    ).lstrip()

    doc_path = tmp_path / "main.yar"
    doc_path.write_text(text)

    provider = DocumentLinksProvider()
    links = provider.get_document_links(text, path_to_uri(doc_path))

    assert len(links) >= 2
    targets = {link.target for link in links}
    assert any("yara.readthedocs.io" in t for t in targets)
    assert any(t.startswith("file://") for t in targets)


def test_document_links_fallback(tmp_path) -> None:
    text = 'import "pe"\ninclude "missing.yar"\n'
    doc_path = tmp_path / "doc.yar"
    doc_path.write_text(text)

    provider = DocumentLinksProvider()
    links = provider._fallback_links(text, path_to_uri(doc_path))

    assert any(link.target and "pe.html" in link.target for link in links)


def test_document_links_helper_edges(tmp_path) -> None:
    include_path = tmp_path / "inc.yar"
    include_path.write_text("rule inc { condition: true }")

    provider = DocumentLinksProvider()
    text = 'import "pe"\ninclude "inc.yar"\n'

    ast_links = provider.get_document_links(text, str(tmp_path / "doc.yar"))
    assert len(ast_links) == 2

    doc = DocumentContext(str(tmp_path / "doc.yar"), text)
    assert doc.get_include_target_uri("inc.yar") is not None
    assert DocumentContext("file://\0bad", "").get_include_target_uri("missing.yar") is None

    fallback_links = provider._fallback_links(
        'import "unknown"\ninclude "missing.yar"\nimport "pe"\n', path_to_uri(tmp_path / "doc.yar")
    )
    assert len(fallback_links) == 1
    assert "pe.html" in (fallback_links[0].target or "")


def test_document_links_parser_fallback_and_error_edges(tmp_path) -> None:
    include_path = tmp_path / "inc.yar"
    include_path.write_text("rule inc { condition: true }")
    provider = DocumentLinksProvider()

    broken_text = 'import "pe"\ninclude "inc.yar"\nrule bad { condition: '
    links = provider.get_document_links(broken_text, path_to_uri(tmp_path / "doc.yar"))
    assert len(links) == 2
    assert DocumentContext("file://\0bad", "").get_include_target_uri("missing.yar") is None


def test_document_links_without_runtime_include_local_rule_links() -> None:
    text = """
rule shared_rule {
  condition:
    true
}

rule main {
  condition:
    shared_rule
}
""".lstrip()

    provider = DocumentLinksProvider()
    links = provider.get_document_links(text, "file://local.yar")

    rule_link = next(link for link in links if link.tooltip == "Go to rule shared_rule")
    assert rule_link.target == "file://local.yar"
    assert rule_link.range.start.line == 7



from __future__ import annotations

import io
import mmap
from pathlib import Path

from yaraast.performance.streaming_parser import StreamingParser


def test_streaming_parser_bytes_stream_remaining_buffer_and_reset(tmp_path: Path) -> None:
    parser = StreamingParser(buffer_size=8)

    raw = b"rule r1 {\n condition:\n true\n}"  # no trailing newline
    rules = list(parser.parse_stream(io.BytesIO(raw)))
    assert len(rules) == 1

    parser.reset_statistics()
    stats = parser.get_statistics()
    assert stats["rules_parsed"] == 0
    assert stats["bytes_processed"] == 0
    assert stats["parse_errors"] == 0
    assert stats["files_processed"] == 0
    assert stats["files_successful"] == 0
    assert stats["total_parse_time"] == 0

    # parse_rules_from_file exception path
    missing = tmp_path / "missing.yar"
    results = list(parser.parse_rules_from_file(missing))
    assert len(results) == 1
    assert results[0].status.name == "ERROR"


def test_streaming_parser_progress_cancel_and_memory_paths(tmp_path: Path) -> None:
    p1 = tmp_path / "a.yar"
    p2 = tmp_path / "b.yar"
    p1.write_text("rule a { condition: true }", encoding="utf-8")
    p2.write_text("rule b { condition: true }", encoding="utf-8")

    progress_calls: list[tuple[int, int, str]] = []
    parser = StreamingParser(
        progress_callback=lambda i, total, path: progress_calls.append((i, total, path))
    )
    rows = list(parser.parse_files([p1, p2]))
    assert len(rows) == 2
    assert progress_calls

    # cancel path for parse_files
    parser2 = StreamingParser()
    parser2.cancel()
    assert list(parser2.parse_files([p1, p2])) == []

    # _memory_limit_exceeded true path with 0MB threshold (current RSS > 0)
    parser3 = StreamingParser(max_memory_mb=0)
    assert parser3._memory_limit_exceeded() is True
    parser3._maybe_collect_garbage()


def test_streaming_parser_mmap_import_include_and_token_branches(tmp_path: Path) -> None:
    code = """
import "pe"
include "common.yar"
private rule r1 {
  condition:
    true
}
global rule r2 : t1 t2 {
  strings:
    $a = "x"
  condition:
    $a
}
"""
    f = tmp_path / "mm.yar"
    f.write_text(code.strip() + "\n", encoding="utf-8")

    parser = StreamingParser()
    rules = list(parser.parse_file(f))
    assert len(rules) == 2
    assert [r.name for r in rules] == ["r1", "r2"]


def test_streaming_parser_chunk_residual_callbacks_and_cancellation(tmp_path: Path) -> None:
    text = """
rule a {
  condition:
    true
}

rule b {
  condition:
    true
}

rule c {
  condition:
    true
}"""
    path = tmp_path / "three.yar"
    path.write_text(text, encoding="utf-8")

    parser = StreamingParser(buffer_size=7)
    chunks = list(parser.parse_file_chunked(path, chunk_size=2))
    assert [len(chunk) for chunk in chunks] == [2, 1]

    callback_names: list[str] = []
    stream_rules = list(
        parser.parse_stream(
            io.StringIO(text), callback=lambda rule: callback_names.append(rule.name)
        )
    )
    assert [r.name for r in stream_rules] == ["a", "b", "c"]
    assert callback_names == ["a", "b", "c"]

    parser2 = StreamingParser()
    parser2.cancel()
    assert list(parser2.parse_rules_from_file(path)) == []


def test_streaming_parser_parse_files_error_memory_except_and_rule_text_failures(
    tmp_path: Path,
) -> None:
    missing = tmp_path / "nope.yar"
    parser = StreamingParser()
    results = list(parser.parse_files([missing]))
    assert len(results) == 1
    assert results[0].status.name == "ERROR"

    parser_bad_mem = StreamingParser(max_memory_mb="bad")  # type: ignore[arg-type]
    assert parser_bad_mem._memory_limit_exceeded() is False

    assert parser._parse_rule_text('import "pe"') is None


def test_streaming_parser_parse_stream_invalid_tail_and_mmap_invalid_forms(tmp_path: Path) -> None:
    parser = StreamingParser(buffer_size=8)
    invalid_tail = "rule tail {\n  condition:\n    true"
    assert list(parser.parse_stream(io.StringIO(invalid_tail))) == []
    assert parser.get_statistics()["parse_errors"] >= 1

    invalid_cases = [
        "private\n",
        "rule { condition: true }\n",
        "rule missing_name\n",
        'import "pe"\n',
        "import\n",
        "rule tagged : tag\n",
    ]
    for idx, content in enumerate(invalid_cases):
        path = tmp_path / f"invalid_{idx}.yar"
        path.write_text(content, encoding="utf-8")
        with path.open("rb") as fh, mmap.mmap(fh.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            assert list(parser._parse_mmap(mm)) == []


def test_streaming_parser_parse_rules_from_file_success_progress_recursive_and_estimate(
    tmp_path: Path,
) -> None:
    root = tmp_path / "root"
    nested = root / "nested"
    nested.mkdir(parents=True)

    file1 = root / "one.yar"
    file2 = nested / "two.yar"
    file1.write_text("rule one { condition: true }", encoding="utf-8")
    file2.write_text("rule two { condition: true }", encoding="utf-8")

    parser = StreamingParser(buffer_size=4)
    results = list(parser.parse_rules_from_file(file1))
    assert len(results) == 1
    assert results[0].status.name == "SUCCESS"
    assert results[0].rule_name == "one"
    assert results[0].rule_count == 1
    assert results[0].ast is not None

    progress_calls: list[tuple[int, int]] = []
    rules = parser.parse_with_progress(
        file1,
        lambda processed, total: progress_calls.append((processed, total)),
    )
    assert [rule.name for rule in rules] == ["one"]
    assert progress_calls

    recursive_results = list(parser.parse_directory(root, recursive=True))
    assert len(recursive_results) == 2
    assert {r.rule_name for r in recursive_results} == {"one", "two"}

    estimate = parser.estimate_memory_usage(file1)
    assert estimate["file_size_mb"] >= 0
    assert estimate["estimated_ast_mb"] >= estimate["file_size_mb"]
    assert estimate["estimated_peak_mb"] >= estimate["estimated_ast_mb"]


def test_streaming_parser_mmap_nested_brace_callback_and_invalid_parse_object(
    tmp_path: Path,
) -> None:
    nested_brace = "rule weird { condition: { true } }"
    path = tmp_path / "nested_brace.yar"
    path.write_text(nested_brace, encoding="utf-8")

    parser = StreamingParser()
    with path.open("rb") as fh, mmap.mmap(fh.fileno(), 0, access=mmap.ACCESS_READ) as mm:
        assert list(parser._parse_mmap(mm, callback=lambda rule: None)) == []
        assert parser.get_statistics()["parse_errors"] >= 1



"""Additional tests for string diagram render helpers without mocks."""

from __future__ import annotations

from yaraast.ast.strings import (
    HexAlternative,
    HexByte,
    HexJump,
    HexString,
    HexWildcard,
    PlainString,
    RegexString,
)
from yaraast.metrics.string_diagrams import (
    StringDiagramGenerator,
    analyze_string_patterns,
    create_hex_diagram,
    create_regex_diagram,
    generate_pattern_report,
    generate_string_diagram,
)


def test_render_mixin_and_convenience_functions() -> None:
    plain = PlainString(identifier="$a", value="foobar", modifiers=["ascii"])
    regex = RegexString(identifier="$b", regex="^(ab)+$", modifiers=["nocase"])
    hexs = HexString(
        identifier="$c",
        tokens=[
            HexByte(value=0x6A),
            HexWildcard(),
            HexJump(min_jump=2, max_jump=2),
            HexAlternative(alternatives=[0x90, 0x91]),
        ],
        modifiers=["wide"],
    )

    gen = StringDiagramGenerator()
    assert "PlainString: $a" in gen.generate(plain)
    assert "RegexString: $b" in gen.generate(regex)
    assert "HexString: $c" in gen.generate(hexs)
    assert "Unknown string type" in gen.generate(object())

    assert "6A ?? [2] (90|91)" in create_hex_diagram(hexs.tokens)
    regex_diag = create_regex_diagram("^[ab]+(cd)?$")
    assert "Pattern:" in regex_diag and "Capture groups" in regex_diag

    conv = generate_string_diagram(plain)
    assert 'Value: "foobar"' in conv


def test_analyze_and_report_patterns() -> None:
    strings = [
        PlainString(identifier="$a", value="abc-common", modifiers=["ascii"]),
        PlainString(identifier="$b", value="abc-other", modifiers=["ascii"]),
        PlainString(identifier="$c", value="abc-common", modifiers=["wide"]),
        RegexString(identifier="$r", regex="ab+", modifiers=[]),
        HexString(identifier="$h", tokens=[HexByte(value=0x41)], modifiers=[]),
    ]

    analysis = analyze_string_patterns(strings)
    assert analysis["total_strings"] == 5
    assert analysis["types"]["plain"] == 3
    assert "abc-" in analysis["patterns"]["common_prefixes"]
    assert "abc-common" in analysis["patterns"]["duplicates"]

    report = generate_pattern_report(strings)
    assert report["summary"]["total"] == 5
    assert len(report["details"]) == 5



"""AST formatting and diff services shared by CLI and LSP."""

from __future__ import annotations

from dataclasses import dataclass, field
import hashlib
import json
from pathlib import Path
from typing import Any

from yaraast.ast.base import ASTNode, YaraFile
from yaraast.ast.conditions import Condition
from yaraast.ast.rules import Rule
from yaraast.ast.strings import HexString, PlainString, RegexString
from yaraast.codegen.generator import CodeGenerator
from yaraast.codegen.pretty_printer import PrettyPrinter
from yaraast.errors import YaraASTError
from yaraast.parser.parser import Parser
from yaraast.visitor.base import BaseVisitor

StringDef = PlainString | HexString | RegexString


@dataclass
class ASTDiffResult:
    """Result of AST-based diff analysis."""

    has_changes: bool
    logical_changes: list[str] = field(default_factory=list)
    structural_changes: list[str] = field(default_factory=list)
    style_only_changes: list[str] = field(default_factory=list)
    added_rules: list[str] = field(default_factory=list)
    removed_rules: list[str] = field(default_factory=list)
    modified_rules: list[str] = field(default_factory=list)
    change_summary: dict[str, int] = field(default_factory=dict)


class ASTStructuralAnalyzer(BaseVisitor[Any]):
    """Analyze AST structure for diffing purposes."""

    def __init__(self) -> None:
        super().__init__()
        self.structural_hash: dict[str, str] = {}
        self.rule_signatures: dict[str, str] = {}
        self.string_signatures: dict[str, str] = {}
        self.condition_signatures: dict[str, str] = {}

    def analyze(self, ast: YaraFile) -> dict[str, Any]:
        self.structural_hash.clear()
        self.rule_signatures.clear()
        self.string_signatures.clear()
        self.condition_signatures.clear()
        self.visit(ast)
        return {
            "structural_hash": self.structural_hash,
            "rule_signatures": self.rule_signatures,
            "string_signatures": self.string_signatures,
            "condition_signatures": self.condition_signatures,
            "total_rules": len(ast.rules),
            "total_imports": len(ast.imports),
            "total_includes": len(ast.includes),
        }

    def visit_yara_file(self, node: YaraFile) -> Any:
        self.structural_hash["file"] = self._hash_dict(
            {
                "imports": [imp.module for imp in node.imports],
                "includes": [inc.path for inc in node.includes],
                "rules": [rule.name for rule in node.rules],
            }
        )
        for rule in node.rules:
            self.visit(rule)
        return super().visit_yara_file(node)

    def visit_rule(self, rule: Rule) -> Any:
        meta_data = getattr(rule, "meta", [])
        meta_keys = sorted([getattr(m, "key", "") for m in meta_data if hasattr(m, "key")])

        self.rule_signatures[rule.name] = self._hash_dict(
            {
                "name": rule.name,
                "modifiers": sorted(str(m) for m in getattr(rule, "modifiers", [])),
                "tags": sorted([tag.name for tag in getattr(rule, "tags", [])]),
                "meta_keys": meta_keys,
                "string_identifiers": sorted([s.identifier for s in getattr(rule, "strings", [])]),
                "has_condition": rule.condition is not None,
            }
        )

        for string_def in getattr(rule, "strings", []):
            self._analyze_string(string_def, rule.name)
        if rule.condition:
            self._analyze_condition(rule.condition, rule.name)
        return super().visit_rule(rule)

    def _analyze_string(self, string_def: StringDef, rule_name: str = "") -> None:
        string_structure = {
            "identifier": string_def.identifier,
            "type": type(string_def).__name__,
            "modifiers": sorted([str(mod) for mod in getattr(string_def, "modifiers", [])]),
        }
        if hasattr(string_def, "value"):
            string_structure["content_type"] = "literal"
            string_structure["content_hash"] = hashlib.md5(
                str(string_def.value).encode(),
                usedforsecurity=False,
            ).hexdigest()
        elif hasattr(string_def, "regex"):
            string_structure["content_type"] = "regex"
            string_structure["content_hash"] = hashlib.md5(
                str(string_def.regex).encode(),
                usedforsecurity=False,
            ).hexdigest()
        elif hasattr(string_def, "tokens"):
            string_structure["content_type"] = "hex"
            string_structure["token_count"] = str(len(getattr(string_def, "tokens", [])))
        sig_key = f"{rule_name}:{string_def.identifier}" if rule_name else string_def.identifier
        self.string_signatures[sig_key] = self._hash_dict(string_structure)

    def _analyze_condition(self, condition: Condition, rule_name: str) -> None:
        self.condition_signatures[f"{rule_name}.condition"] = self._hash_dict(
            self._get_condition_structure(condition)
        )

    def _get_condition_structure(self, node: ASTNode) -> dict[str, Any]:
        if node is None:
            return {"type": "empty"}
        structure: dict[str, Any] = {"type": type(node).__name__}
        if hasattr(node, "operator"):
            structure["operator"] = node.operator
        if hasattr(node, "left") and hasattr(node, "right"):
            structure["left"] = self._get_condition_structure(node.left)
            structure["right"] = self._get_condition_structure(node.right)
        elif hasattr(node, "operand"):
            structure["operand"] = self._get_condition_structure(node.operand)
        elif hasattr(node, "children"):
            structure["children"] = [
                self._get_condition_structure(child) for child in node.children()
            ]
        return structure

    def _hash_dict(self, data: dict[str, Any]) -> str:
        payload = json.dumps(data, sort_keys=True, default=str, separators=(",", ":"))
        return hashlib.md5(payload.encode(), usedforsecurity=False).hexdigest()


class ASTDiffer:
    """Compare ASTs and identify logical vs stylistic changes."""

    def diff_files(self, file1_path: Path, file2_path: Path) -> ASTDiffResult:
        try:
            with Path(file1_path).open() as f:
                content1 = f.read()
                ast1 = Parser(content1).parse()
            with Path(file2_path).open() as f:
                content2 = f.read()
                ast2 = Parser(content2).parse()
            result = self.diff_asts(ast1, ast2)
            return self._detect_style_changes_from_text(content1, content2, result)
        except (YaraASTError, OSError) as exc:  # file I/O + parse errors from multiple paths
            result = ASTDiffResult(has_changes=True)
            result.logical_changes.append(f"Error comparing files: {exc}")
            return result

    def diff_asts(self, ast1: YaraFile, ast2: YaraFile) -> ASTDiffResult:
        analysis1 = ASTStructuralAnalyzer().analyze(ast1)
        analysis2 = ASTStructuralAnalyzer().analyze(ast2)
        result = ASTDiffResult(has_changes=False)

        if analysis1["structural_hash"]["file"] != analysis2["structural_hash"]["file"]:
            result.structural_changes.append("File structure changed (imports/includes/rule order)")
            result.has_changes = True

        rules1 = set(analysis1["rule_signatures"].keys())
        rules2 = set(analysis2["rule_signatures"].keys())
        result.added_rules = list(rules2 - rules1)
        result.removed_rules = list(rules1 - rules2)
        result.modified_rules = []

        for rule_name in rules1 & rules2:
            if analysis1["rule_signatures"][rule_name] != analysis2["rule_signatures"][rule_name]:
                result.modified_rules.append(rule_name)
                result.logical_changes.append(
                    f"Rule '{rule_name}' modified (logic/structure changed)"
                )

        strings1 = set(analysis1["string_signatures"].keys())
        strings2 = set(analysis2["string_signatures"].keys())
        added_strings = strings2 - strings1
        removed_strings = strings1 - strings2
        if added_strings:
            result.logical_changes.append(f"Added strings: {', '.join(added_strings)}")
        if removed_strings:
            result.logical_changes.append(f"Removed strings: {', '.join(removed_strings)}")

        for string_id in strings1 & strings2:
            if (
                analysis1["string_signatures"][string_id]
                != analysis2["string_signatures"][string_id]
            ):
                result.logical_changes.append(f"String '{string_id}' content modified")

        conditions1 = analysis1["condition_signatures"]
        conditions2 = analysis2["condition_signatures"]
        for condition_name in set(conditions1) & set(conditions2):
            if conditions1[condition_name] != conditions2[condition_name]:
                rule_name = condition_name.removesuffix(".condition")
                result.logical_changes.append(f"Condition logic changed in rule '{rule_name}'")

        result.has_changes = bool(
            result.logical_changes
            or result.structural_changes
            or result.added_rules
            or result.removed_rules
        )
        result.change_summary = {
            "logical_changes": len(result.logical_changes),
            "structural_changes": len(result.structural_changes),
            "added_rules": len(result.added_rules),
            "removed_rules": len(result.removed_rules),
            "modified_rules": len(result.modified_rules),
            "style_only_changes": len(result.style_only_changes),
            "style_changes": len(result.style_only_changes),
        }
        return result

    def _detect_style_changes_from_text(
        self,
        content1: str,
        content2: str,
        result: ASTDiffResult,
    ) -> ASTDiffResult:
        result.change_summary.setdefault("style_changes", 0)
        result.change_summary.setdefault("style_only_changes", 0)
        if content1 != content2 and not (
            result.logical_changes
            or result.structural_changes
            or result.added_rules
            or result.removed_rules
        ):
            result.has_changes = True
            result.style_only_changes.append("spacing/formatting or whitespace/indentation changed")
            result.change_summary["style_only_changes"] = len(result.style_only_changes)
            result.change_summary["style_changes"] = len(result.style_only_changes)
        return result

    def _detect_style_changes(
        self,
        ast1: YaraFile,
        ast2: YaraFile,
        result: ASTDiffResult,
    ) -> ASTDiffResult:
        """Backward-compatible hook retained for existing tests."""
        if ast1 is ast2:
            result.change_summary.setdefault("style_changes", len(result.style_only_changes))
            return result
        return result


class ASTFormatter:
    """AST-based code formatter using CodeGenerator."""

    def __init__(self) -> None:
        self.generator = CodeGenerator()
        self.pretty_printer = PrettyPrinter()

    def format_file(
        self,
        input_path: Path,
        output_path: Path | None = None,
        style: str = "default",
    ) -> tuple[bool, str]:
        try:
            with Path(input_path).open() as f:
                ast = Parser(f.read()).parse()
            formatted = self.format_ast(ast, style)
            if output_path:
                with Path(output_path).open("w") as f:
                    f.write(formatted)
                return True, f"Formatted file written to {output_path}"
            return True, formatted
        except (YaraASTError, OSError) as exc:  # file I/O + parse + codegen errors
            return False, f"Formatting error: {exc}"

    def format_ast(self, ast: YaraFile, style: str = "default") -> str:
        if style == "compact":
            return self.generator.generate(ast)
        if style in ("pretty", "verbose"):
            self.pretty_printer = PrettyPrinter()
            return self.pretty_printer.pretty_print(ast)
        return self.pretty_printer.pretty_print(ast)

    def check_format(self, file_path: Path) -> tuple[bool, list[str]]:
        try:
            with Path(file_path).open() as f:
                original = f.read()
            formatted = self.pretty_printer.pretty_print(Parser(original).parse())
            if original.strip() == formatted.strip():
                return False, []
            issues = []
            for i, (orig, fmt) in enumerate(
                zip(original.strip().split("\n"), formatted.strip().split("\n"), strict=False),
                1,
            ):
                if orig != fmt:
                    issues.append(f"Line {i}: formatting issue")
            return len(issues) > 0, issues
        except (YaraASTError, OSError) as exc:  # file I/O + parse + codegen errors
            return False, [f"Check error: {exc}"]



"""Tests for batch processor utilities (no mocks)."""

from __future__ import annotations

from pathlib import Path
from textwrap import dedent

from yaraast.performance.batch_processor import BatchOperation, BatchProcessor


def _write_rules(tmp_path: Path) -> list[Path]:
    code1 = """
    rule batch_one {
        strings:
            $a = "abc"
        condition:
            $a
    }
    """
    code2 = """
    rule batch_two {
        condition:
            true
    }
    """
    p1 = tmp_path / "r1.yar"
    p2 = tmp_path / "r2.yar"
    p1.write_text(dedent(code1), encoding="utf-8")
    p2.write_text(dedent(code2), encoding="utf-8")
    return [p1, p2]


def test_batch_processor_process_files_and_directory(tmp_path: Path) -> None:
    paths = _write_rules(tmp_path)
    processor = BatchProcessor(batch_size=1)

    result = processor.process_files(paths, BatchOperation.HTML_TREE, output_dir=tmp_path / "out")
    assert result.successful_count == 2
    assert result.output_files

    result2 = processor.process_directory(
        tmp_path, BatchOperation.SERIALIZE, output_dir=tmp_path / "jsons"
    )
    assert result2.successful_count == 2
    assert result2.output_files


def test_batch_processor_process_batch_and_stats(tmp_path: Path) -> None:
    paths = _write_rules(tmp_path)
    processor = BatchProcessor(batch_size=1)

    results = processor.process_batch(paths, BatchOperation.PARSE)
    assert len(results) == 2

    stats = processor.get_statistics()
    assert stats["items_processed"] >= 2

    processor.reset_statistics()
    stats_after = processor.get_statistics()
    assert stats_after["items_processed"] == 0


def test_batch_processor_large_file_split(tmp_path: Path) -> None:
    code = """
    rule r1 { condition: true }
    rule r2 { condition: true }
    """
    path = tmp_path / "many.yar"
    path.write_text(dedent(code), encoding="utf-8")

    processor = BatchProcessor()
    results = processor.process_large_file(
        path,
        operations=[BatchOperation.PARSE, BatchOperation.COMPLEXITY],
        output_dir=tmp_path,
        split_rules=True,
    )

    assert results[BatchOperation.PARSE].successful_count == 2
    assert results[BatchOperation.COMPLEXITY].summary



"""More tests for memory optimizer (no mocks)."""

from __future__ import annotations

from textwrap import dedent

from yaraast.parser import Parser
from yaraast.performance.memory_optimizer import MemoryOptimizer


def _parse_yara(code: str):
    parser = Parser()
    return parser.parse(dedent(code))


def test_memory_optimizer_basic_and_stats() -> None:
    code = """
    import "pe"
    rule mem_rule : tag1 {
        strings:
            $a = "abc"
            $b = "abc"
        condition:
            $a and pe.number_of_sections > 0
    }
    """
    ast = _parse_yara(code)
    optimizer = MemoryOptimizer(enable_tracking=True)

    optimizer.track_object(ast)
    stats_before = optimizer.get_statistics()
    assert stats_before["total_objects"] >= 1

    optimized = optimizer.optimize(ast)
    assert optimized is not ast  # optimizer returns a new copy, not the original

    stats = optimizer.get_statistics()
    assert stats["nodes_processed"] >= 1
    assert stats["string_pool_size"] >= 1

    mem_stats = optimizer.get_memory_stats()
    assert mem_stats.total_objects >= 1

    optimizer.clear_caches()
    stats_after = optimizer.get_statistics()
    assert stats_after["string_pool_size"] == 0


def test_memory_optimizer_context_and_cleanup() -> None:
    optimizer = MemoryOptimizer(enable_tracking=True)
    with optimizer.memory_managed_context():
        optimizer.track_object({"x": 1})

    mem_stats = optimizer.get_memory_stats()
    assert mem_stats.total_objects >= 0

    cleaned = optimizer.force_cleanup()
    assert isinstance(cleaned, int)


def test_memory_optimizer_recommendations_and_pool() -> None:
    optimizer = MemoryOptimizer()
    rec_small = optimizer.optimize_for_large_collection(10)
    rec_medium = optimizer.optimize_for_large_collection(200)
    rec_large = optimizer.optimize_for_large_collection(2000)

    assert rec_small["batch_size"] <= rec_medium["batch_size"]
    assert rec_large["use_streaming"] is True
    assert rec_large["enable_pooling"] is True

    ast1 = optimizer.create_memory_efficient_ast()
    optimizer.return_ast_to_pool(ast1)
    ast2 = optimizer.create_memory_efficient_ast()
    assert ast2 is ast1



"""Enhanced fluent rule builder with comprehensive chaining support."""

from __future__ import annotations

from typing import TYPE_CHECKING, Self

from yaraast.ast.rules import Rule
from yaraast.builder.fluent_condition_builder import FluentConditionBuilder
from yaraast.builder.fluent_rule_helpers import apply_last_string_modifier, combine_condition
from yaraast.builder.fluent_string_builder import FluentStringBuilder
from yaraast.builder.rule_builder import RuleBuilder
from yaraast.errors import ValidationError

if TYPE_CHECKING:
    from collections.abc import Callable

    from yaraast.ast.base import YaraFile
    from yaraast.ast.expressions import Expression
    from yaraast.builder.fluent_file_builder import FluentYaraFileBuilder


class FluentRuleBuilder:
    """Enhanced fluent builder for YARA rules with comprehensive chaining."""

    # Constants
    MZ_HEADER = "$mz"
    YARA_AST_STR = "YARA AST"

    def __init__(self, name: str | None = None) -> None:
        self._rule_builder = RuleBuilder()
        if name:
            self._rule_builder.with_name(name)
        self._string_builders: list[FluentStringBuilder] = []

    # Rule metadata methods
    def named(self, name: str) -> Self:
        """Set the rule name."""
        self._rule_builder.with_name(name)
        return self

    def private(self) -> Self:
        """Mark rule as private."""
        self._rule_builder.private()
        return self

    def global_(self) -> Self:
        """Mark rule as global."""
        self._rule_builder.global_()
        return self

    def public(self) -> Self:
        """Mark rule as public (remove private modifier)."""
        self._rule_builder.public()
        return self

    def tagged(self, *tags: str) -> Self:
        """Add tags to the rule."""
        self._rule_builder.with_tags(*tags)
        return self

    def with_tag(self, tag: str) -> Self:
        """Add a single tag."""
        self._rule_builder.with_tag(tag)
        return self

    def meta(self, key: str, value: str | int | bool) -> Self:
        """Add metadata."""
        self._rule_builder.with_meta(key, value)
        return self

    def with_meta(self, key: str, value: str | int | bool) -> Self:
        """Add metadata (alias for meta)."""
        return self.meta(key, value)

    def authored_by(self, author: str) -> Self:
        """Set author."""
        self._rule_builder.with_author(author)
        return self

    def described_as(self, description: str) -> Self:
        """Set description."""
        self._rule_builder.with_description(description)
        return self

    def versioned(self, version: int) -> Self:
        """Set version."""
        self._rule_builder.with_version(version)
        return self

    # String definition methods
    def with_string(
        self,
        identifier_or_builder: str | FluentStringBuilder,
        value: str | None = None,
    ) -> Self:
        """Add a string using FluentStringBuilder or simple string."""
        if isinstance(identifier_or_builder, str) and value is not None:
            # Simple string case
            builder = FluentStringBuilder.text_string(identifier_or_builder, value)
            self._string_builders.append(builder)
        elif isinstance(identifier_or_builder, FluentStringBuilder):
            # Builder case
            self._string_builders.append(identifier_or_builder)
        else:
            msg = "Either provide (identifier, value) or a FluentStringBuilder"
            raise ValidationError(
                msg,
            )
        return self

    def string(self, identifier: str) -> FluentStringContext:
        """Start defining a string (returns context for chaining)."""
        return FluentStringContext(self, identifier)

    def text_string(self, identifier: str, content: str) -> Self:
        """Add a text string."""
        builder = FluentStringBuilder.text_string(identifier, content)
        return self.with_string(builder)

    def hex_string(self, identifier: str, pattern: str) -> Self:
        """Add a hex string."""
        builder = FluentStringBuilder.hex_string(identifier, pattern)
        return self.with_string(builder)

    def regex_string(self, identifier: str, pattern: str) -> Self:
        """Add a regex string."""
        builder = FluentStringBuilder.regex_string(identifier, pattern)
        return self.with_string(builder)

    # Common string patterns
    def with_mz_header(self, identifier: str = "$mz") -> Self:
        """Add MZ header string."""
        return self.with_string(FluentStringBuilder.string(identifier).mz_header())

    def mz_header(self, identifier: str = "$mz") -> Self:
        """Add MZ header string (alias for with_mz_header)."""
        return self.with_mz_header(identifier)

    def pe_header(self, identifier: str = "$pe") -> Self:
        """Add PE header string."""
        return self.with_string(FluentStringBuilder.string(identifier).pe_header())

    def elf_header(self, identifier: str = "$elf") -> Self:
        """Add ELF header string."""
        return self.with_string(FluentStringBuilder.string(identifier).elf_header())

    def email_pattern(self, identifier: str = "$email") -> Self:
        """Add email regex pattern."""
        return self.with_string(FluentStringBuilder.string(identifier).email_pattern())

    def ip_pattern(self, identifier: str = "$ip") -> Self:
        """Add IP address regex pattern."""
        return self.with_string(
            FluentStringBuilder.string(identifier).ip_address_pattern(),
        )

    def url_pattern(self, identifier: str = "$url") -> Self:
        """Add URL regex pattern."""
        return self.with_string(FluentStringBuilder.string(identifier).url_pattern())

    # String modifier methods (apply to most recent string)
    def nocase(self) -> Self:
        """Add nocase modifier to the most recent string."""
        apply_last_string_modifier(self._string_builders, "nocase")
        return self

    def ascii(self) -> Self:
        """Add ASCII modifier to the most recent string."""
        apply_last_string_modifier(self._string_builders, "ascii")
        return self

    def wide(self) -> Self:
        """Add wide modifier to the most recent string."""
        apply_last_string_modifier(self._string_builders, "wide")
        return self

    def fullword(self) -> Self:
        """Add fullword modifier to the most recent string."""
        apply_last_string_modifier(self._string_builders, "fullword")
        return self

    def private_string(self) -> Self:
        """Add private modifier to the most recent string."""
        apply_last_string_modifier(self._string_builders, "private")
        return self

    def xor(self, key: int | str | None = None) -> Self:
        """Add XOR modifier to the most recent string."""
        apply_last_string_modifier(self._string_builders, "xor", key)
        return self

    def base64(self) -> Self:
        """Add base64 modifier to the most recent string."""
        apply_last_string_modifier(self._string_builders, "base64")
        return self

    # Condition methods
    def condition(self, condition: str | Expression | FluentConditionBuilder) -> Self:
        """Set the rule condition."""
        if isinstance(condition, FluentConditionBuilder):
            self._rule_builder.with_condition(condition.build())
        else:
            self._rule_builder.with_condition(condition)
        return self

    def when(self, condition: str | Expression | FluentConditionBuilder) -> Self:
        """Alias for condition."""
        return self.condition(condition)

    def with_condition(
        self,
        condition: str | Expression | FluentConditionBuilder,
    ) -> Self:
        """Alias for condition."""
        return self.condition(condition)

    def matches_any(self) -> Self:
        """Condition: any of them."""
        return self.condition("any of them")

    def matches_all(self) -> Self:
        """Condition: all of them."""
        return self.condition("all of them")

    def matches_one_of(self, *strings: str) -> Self:
        """Condition: one of specified strings."""
        builder = FluentConditionBuilder().one_of(*strings)
        return self.condition(builder)

    def matches_any_of(self, *strings: str) -> Self:
        """Condition: any of specified strings."""
        builder = FluentConditionBuilder().any_of(*strings)
        return self.condition(builder)

    def matches_all_of(self, *strings: str) -> Self:
        """Condition: all of specified strings."""
        builder = FluentConditionBuilder().all_of(*strings)
        return self.condition(builder)

    def with_condition_builder(
        self,
        builder_func: Callable[[FluentConditionBuilder], FluentConditionBuilder],
    ) -> Self:
        """Set condition using a builder function."""
        condition_builder = FluentConditionBuilder()
        result = builder_func(condition_builder)
        return self.condition(result)

    # File property conditions
    def for_small_files(self) -> Self:
        """Add small file condition (< 1MB)."""
        condition_builder = FluentConditionBuilder().small_file()
        combined = combine_condition(self._rule_builder.get_condition(), condition_builder)
        return self.condition(combined)

    def for_large_files(self) -> Self:
        """Add large file condition (> 10MB)."""
        condition_builder = FluentConditionBuilder().large_file()
        combined = combine_condition(self._rule_builder.get_condition(), condition_builder)
        return self.condition(combined)

    def for_pe_files(self) -> Self:
        """Add PE file conditions."""
        # Add MZ header if not already present
        if not any(s.identifier in ["$mz", self.MZ_HEADER] for s in self._string_builders):
            self.with_mz_header(self.MZ_HEADER)

        condition_builder = FluentConditionBuilder().string_matches(self.MZ_HEADER).at(0)
        combined = combine_condition(self._rule_builder.get_condition(), condition_builder)
        return self.condition(combined)

    def for_executables(self) -> Self:
        """Add executable file conditions."""
        return self.for_pe_files().for_large_files()

    # Build method
    def build(self) -> Rule:
        """Build the rule."""
        self._rule_builder.add_string_definitions(
            [string_builder.build() for string_builder in self._string_builders],
        )

        rule = self._rule_builder.build()
        return rule


class FluentStringContext:
    """Context for fluent string building within a rule."""

    def __init__(self, rule_builder: FluentRuleBuilder, identifier: str) -> None:
        self.rule_builder = rule_builder
        self.string_builder = FluentStringBuilder(identifier)

    # String content methods
    def literal(self, content: str) -> FluentStringContext:
        """Set string content as literal."""
        self.string_builder.literal(content)
        return self

    def text(self, content: str) -> FluentStringContext:
        """Set string content as text."""
        self.string_builder.text(content)
        return self

    def hex(self, pattern: str) -> FluentStringContext:
        """Set string content as hex pattern."""
        self.string_builder.hex(pattern)
        return self

    def regex(self, pattern: str) -> FluentStringContext:
        """Set string content as regex."""
        self.string_builder.regex(pattern)
        return self

    # String modifiers
    def ascii(self) -> FluentStringContext:
        """Add ASCII modifier."""
        self.string_builder.ascii()
        return self

    def wide(self) -> FluentStringContext:
        """Add wide modifier."""
        self.string_builder.wide()
        return self

    def nocase(self) -> FluentStringContext:
        """Add nocase modifier."""
        self.string_builder.nocase()
        return self

    def fullword(self) -> FluentStringContext:
        """Add fullword modifier."""
        self.string_builder.fullword()
        return self

    def private(self) -> FluentStringContext:
        """Add private modifier."""
        self.string_builder.private()
        return self

    def xor(self, key: int | str | None = None) -> FluentStringContext:
        """Add XOR modifier."""
        self.string_builder.xor(key)
        return self

    def base64(self) -> FluentStringContext:
        """Add base64 modifier."""
        self.string_builder.base64()
        return self

    # Pattern helpers
    def with_mz_header_string(self) -> FluentStringContext:
        """Set as MZ header pattern."""
        self.string_builder.mz_header()
        return self

    def pe_header(self) -> FluentStringContext:
        """Set as PE header pattern."""
        self.string_builder.pe_header()
        return self

    def email_pattern(self) -> FluentStringContext:
        """Set as email regex pattern."""
        self.string_builder.email_pattern()
        return self

    # Return to rule builder
    def then(self) -> FluentRuleBuilder:
        """Return to rule builder after defining string."""
        self.rule_builder._string_builders.append(self.string_builder)
        return self.rule_builder

    def and_string(self, identifier: str) -> FluentStringContext:
        """Add this string and start another."""
        self.rule_builder._string_builders.append(self.string_builder)
        return FluentStringContext(self.rule_builder, identifier)


class FluentRuleBuilderWithFile(FluentRuleBuilder):
    """Rule builder that can return to file builder."""

    def __init__(self, file_builder: FluentYaraFileBuilder, name: str) -> None:
        super().__init__(name)
        self.file_builder = file_builder

    def then_rule(self, name: str) -> FluentRuleBuilderWithFile:
        """Add this rule and start another."""
        rule = self.build()
        self.file_builder.with_rule(rule)
        return FluentRuleBuilderWithFile(self.file_builder, name)

    def then_build_file(self) -> YaraFile:
        """Add this rule and build the file."""
        rule = self.build()
        self.file_builder.with_rule(rule)
        return self.file_builder.build()



"""Real integration tests for CLI parse services without test doubles."""

from __future__ import annotations

from yaraast.cli.parse_services import parse_content_by_dialect


def test_parse_services_auto_yaral() -> None:
    msgs: list[str] = []
    code = """
rule login_events {
  events:
    $e.metadata.event_type = "USER_LOGIN"
  condition:
    $e
}
"""
    ast, lex, par = parse_content_by_dialect(code, "auto", show_status=True, status_cb=msgs.append)
    assert ast is not None
    assert lex == []
    assert par == []
    assert any("Detected dialect: YARA_L" in m for m in msgs)


def test_parse_services_auto_yara_valid() -> None:
    code = """
rule r {
  strings:
    $a = "x"
  condition:
    $a
}
"""
    ast, lex, par = parse_content_by_dialect(code, "auto", show_status=False)
    assert ast is not None
    assert lex == []
    assert par == []


def test_parse_services_auto_yara_invalid_fallback() -> None:
    ast, lex, par = parse_content_by_dialect("rule broken { condition:", "auto", show_status=False)
    assert ast is not None
    assert isinstance(lex, list)
    assert isinstance(par, list)


def test_parse_services_yaral_explicit_and_standard() -> None:
    msgs: list[str] = []
    yaral_code = """
rule ev {
  events:
    $e.metadata.event_type = "X"
  condition:
    $e
}
"""
    ast_yaral, lex_y, par_y = parse_content_by_dialect(
        yaral_code, "yara-l", show_status=True, status_cb=msgs.append
    )
    assert ast_yaral is not None
    assert lex_y == []
    assert par_y == []
    assert any("Using YARA-L parser" in m for m in msgs)

    yara_code = """
rule std {
  condition:
    true
}
"""
    ast_std, lex_s, par_s = parse_content_by_dialect(yara_code, "yara", show_status=False)
    assert ast_std is not None
    assert lex_s == []
    assert par_s == []



"""Additional real coverage for dependency graph graphviz/utils helpers."""

from __future__ import annotations

from pathlib import Path

import graphviz

from yaraast.ast.base import YaraFile
from yaraast.ast.expressions import Identifier
from yaraast.ast.rules import Rule
from yaraast.metrics.dependency_graph_graphviz import (
    add_edge,
    add_import_cluster,
    add_include_cluster,
    add_module_edges,
    add_rules_cluster,
    add_string_reference_edges,
    create_graph,
)
from yaraast.metrics.dependency_graph_helpers import render_graph
from yaraast.metrics.dependency_graph_utils import (
    DependencyGraph,
    build_dependency_graph,
    find_circular_dependencies,
    get_dependency_order,
)


def test_graphviz_helper_empty_paths_and_edge_styles() -> None:
    dot = create_graph("deps", "LR")

    add_import_cluster(dot, [])
    add_include_cluster(dot, [])
    add_rules_cluster(dot, {}, lambda *_: "x", lambda *_: "white")
    add_string_reference_edges(dot, {"rule_a": set()})
    add_module_edges(dot, {"rule_a": {"pe", "math"}}, {"pe"})
    add_edge(dot, "a", "b", style="dashed")

    source = dot.source
    assert "cluster_imports" not in source
    assert "cluster_includes" not in source
    assert "cluster_rules" not in source
    assert "rule_a_strings" not in source
    assert "mod_pe" in source
    assert "mod_math" not in source
    assert "style=dashed" in source

    with_includes = create_graph("deps2", "LR")
    add_include_cluster(with_includes, ["common.yar"])
    include_source = with_includes.source
    assert "cluster_includes" in include_source
    assert "include_common.yar" in include_source


def test_dependency_utils_remaining_paths() -> None:
    ast = YaraFile(
        rules=[
            Rule(name="a", condition=Identifier("b")),
            Rule(name="b", condition=Identifier("c")),
            Rule(name="c", condition=Identifier("a")),
            Rule(name="d", condition=None),
        ]
    )

    graph = build_dependency_graph(ast)
    assert graph.has_node("d")
    assert graph.get_dependencies("d") == set()

    cycles = find_circular_dependencies(graph)
    assert len(cycles) == 1
    assert set(cycles[0]) == {"a", "b", "c"}

    order = get_dependency_order(graph)
    assert set(order) == {"a", "b", "c", "d"}

    acyclic = DependencyGraph()
    acyclic.add_edge("top", "mid")
    acyclic.add_edge("mid", "leaf")
    acyclic.add_node("solo")
    acyclic_order = get_dependency_order(acyclic)
    assert set(acyclic_order) == {"top", "mid", "leaf", "solo"}
    assert acyclic_order.index("leaf") < acyclic_order.index("mid") < acyclic_order.index("top")

    empty = DependencyGraph()
    assert get_dependency_order(empty) == []


def test_render_graph_success_path_with_real_graphviz(tmp_path: Path) -> None:
    dot = graphviz.Digraph(comment="deps")
    dot.node("a")
    dot.node("b")
    dot.edge("a", "b")

    out = tmp_path / "deps.svg"
    result = render_graph(dot, str(out), "svg")

    assert result.endswith(".svg")
    assert Path(result).exists()



"""String parsing helpers."""

from __future__ import annotations

from yaraast.ast.modifiers import StringModifier
from yaraast.ast.strings import HexString, HexToken, PlainString, RegexString, StringDefinition
from yaraast.lexer import TokenType
from yaraast.parser.hex_parser import HexParseError, HexStringParser

from ._shared import ParserError


class StringParsingMixin:
    """Mixin with string parsing helpers."""

    def _parse_strings_section(self) -> list[StringDefinition]:
        """Parse strings section."""
        strings: list[StringDefinition] = []
        anonymous_counter = 0

        while not self._check_any(TokenType.CONDITION, TokenType.RBRACE):
            if not self._check(TokenType.STRING_IDENTIFIER):
                break

            start_token = self._advance()
            identifier = start_token.value

            # Handle anonymous strings (just "$")
            if identifier == "$":
                anonymous_counter += 1
                identifier = f"$anon_{anonymous_counter}"

            if not self._match(TokenType.ASSIGN):
                msg = "Expected '=' after string identifier"
                raise ParserError(msg, self._peek())

            # Parse string value
            if self._match(TokenType.STRING):
                value = self._previous().value
                modifiers = self._parse_string_modifiers()
                strings.append(
                    self._set_node_location_from_tokens(
                        PlainString(identifier=identifier, value=value, modifiers=modifiers),
                        start_token,
                        self._previous(),
                    )
                )
            elif self._match(TokenType.HEX_STRING):
                hex_value = self._previous().value
                tokens = self._parse_hex_string(hex_value)
                modifiers = self._parse_string_modifiers()
                strings.append(
                    self._set_node_location_from_tokens(
                        HexString(identifier=identifier, tokens=tokens, modifiers=modifiers),
                        start_token,
                        self._previous(),
                    )
                )
            elif self._match(TokenType.REGEX):
                regex = self._previous().value
                modifiers = self._parse_string_modifiers()
                strings.append(
                    self._set_node_location_from_tokens(
                        RegexString(identifier=identifier, regex=regex, modifiers=modifiers),
                        start_token,
                        self._previous(),
                    )
                )
            else:
                msg = "Invalid string value"
                raise ParserError(msg, self._peek())

        return strings

    def _parse_string_modifiers(self) -> list[StringModifier]:
        """Parse string modifiers."""
        modifiers: list[StringModifier] = []

        while self._check_any(
            TokenType.NOCASE,
            TokenType.WIDE,
            TokenType.ASCII,
            TokenType.XOR_MOD,
            TokenType.BASE64,
            TokenType.BASE64WIDE,
            TokenType.FULLWORD,
            TokenType.PRIVATE,
        ):
            mod_token = self._advance()
            mod_name = mod_token.value.lower()

            # Some modifiers can have parameters
            if mod_name in ("xor", "base64", "base64wide") and self._match(TokenType.LPAREN):
                # Parse modifier parameters
                value = None
                if mod_name == "xor":
                    # xor takes integer or range: xor(0x01-0xff)
                    if self._match(TokenType.INTEGER):
                        min_val = self._previous().value
                        if self._match(TokenType.MINUS):
                            if self._match(TokenType.INTEGER):
                                max_val = self._previous().value
                                value = (min_val, max_val)
                            else:
                                msg = "Expected integer after '-'"
                                raise ParserError(msg, self._peek())
                        else:
                            value = min_val
                    else:
                        msg = "Expected integer or range in xor"
                        raise ParserError(msg, self._peek())
                else:
                    # base64/base64wide takes optional custom alphabet string
                    if self._match(TokenType.STRING):
                        value = self._previous().value

                if not self._match(TokenType.RPAREN):
                    msg = f"Expected ')' after {mod_name} parameter"
                    raise ParserError(msg, self._peek())

                modifiers.append(StringModifier.from_name_value(mod_name, value))
            else:
                modifiers.append(StringModifier.from_name_value(mod_name))

        return modifiers

    def _parse_hex_string(self, hex_content: str) -> list[HexToken]:
        """Parse hex string content into tokens.

        Delegates to HexStringParser for the actual parsing logic.
        Converts HexParseError to ParserError for consistent error handling.
        """
        try:
            hex_parser = HexStringParser(error_token=self._peek())
            return hex_parser.parse(hex_content)
        except HexParseError as e:
            raise ParserError(str(e), self._peek()) from e



"""Range helpers for LSP document symbols."""

from __future__ import annotations

from typing import Any

from lsprotocol.types import Position, Range

from yaraast.ast.base import Location
from yaraast.lsp.structure import find_quoted_value_range, make_range


def node_range(node: Any, source_text: str) -> Range | None:
    location = getattr(node, "location", None)
    if location is None:
        return None
    return location_to_symbol_range(location, source_text)


def node_value_range(node: Any, source_text: str, value: str) -> Range | None:
    base_range = node_range(node, source_text)
    if base_range is None:
        return None
    return narrow_range_to_value(source_text, base_range, value) or base_range


def quoted_value_range_from_node_line(lines: list[str], node: Any, value: str) -> Range | None:
    location = getattr(node, "location", None)
    if location is None:
        return None
    return find_quoted_value_range(lines, location.line - 1, value)


def narrow_range_to_value(source_text: str, base_range: Range, value: str) -> Range | None:
    lines = source_text.split("\n")
    start_line = base_range.start.line
    end_line = min(base_range.end.line, len(lines) - 1)
    for line_num in range(start_line, end_line + 1):
        line = lines[line_num]
        start_char = base_range.start.character if line_num == start_line else 0
        end_char = base_range.end.character if line_num == end_line else len(line)
        segment = line[start_char:end_char]
        offset = segment.find(value)
        if offset >= 0:
            return make_range(line_num, start_char + offset, start_char + offset + len(value))
        quoted = segment.find(f'"{value}"')
        if quoted >= 0:
            return make_range(
                line_num, start_char + quoted + 1, start_char + quoted + 1 + len(value)
            )
    return None


def location_to_symbol_range(location: Location, source_text: str) -> Range:
    start_line = location.line - 1
    start_character = max(0, location.column - 1)
    if location.end_line is not None and location.end_column is not None:
        return Range(
            start=Position(line=start_line, character=start_character),
            end=Position(
                line=location.end_line - 1,
                character=max(start_character + 1, location.end_column),
            ),
        )
    lines = source_text.split("\n")
    line_text = lines[start_line] if 0 <= start_line < len(lines) else ""
    return Range(
        start=Position(line=start_line, character=start_character),
        end=Position(line=start_line, character=min(len(line_text), start_character + 1)),
    )



"""Service helpers for roundtrip CLI (logic without IO)."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any

from yaraast.ast.base import YaraFile
from yaraast.cli.utils import read_text
from yaraast.codegen.pretty_printer import PrettyPrinter, StylePresets
from yaraast.parser.parser import Parser
from yaraast.serialization.roundtrip_serializer import (
    RoundTripSerializer,
    create_rules_manifest,
    roundtrip_yara,
    serialize_for_pipeline,
)


def serialize_roundtrip_file(
    input_file: Path, format: str, preserve_comments: bool, preserve_formatting: bool
) -> tuple[Any, str]:
    yara_source = read_text(input_file)
    serializer = RoundTripSerializer(
        preserve_comments=preserve_comments,
        preserve_formatting=preserve_formatting,
    )
    return serializer.parse_and_serialize(
        yara_source,
        source_file=str(input_file),
        format=format,
    )


def deserialize_roundtrip_file(
    input_file: Path,
    format: str,
    preserve_formatting: bool,
) -> tuple[Any, str]:
    serializer = RoundTripSerializer()
    serialized = read_text(input_file)
    return serializer.deserialize_and_generate(
        serialized,
        format=format,
        preserve_original_formatting=preserve_formatting,
    )


def test_roundtrip_file(input_file: Path, format: str) -> dict[str, Any]:
    yara_source = read_text(input_file)
    return roundtrip_yara(yara_source, format=format)


def pretty_print_file(
    input_file: Path,
    style: str,
    indent_size: int,
    max_line_length: int,
    align_strings: bool,
    align_meta: bool,
    sort_imports: bool,
    sort_tags: bool,
) -> tuple[Any, str]:
    parser = Parser()
    yara_source = read_text(input_file)
    ast = parser.parse(yara_source)
    if style == "compact":
        options = StylePresets.compact()
    elif style == "dense":
        options = StylePresets.dense()
    elif style == "verbose":
        options = StylePresets.verbose()
    else:
        options = StylePresets.readable()
    options.indent_size = indent_size
    options.max_line_length = max_line_length
    options.align_string_definitions = align_strings
    options.align_meta_values = align_meta
    options.sort_imports = sort_imports
    options.sort_tags = sort_tags
    printer = PrettyPrinter(options)
    formatted_code = printer.pretty_print(ast)
    return ast, formatted_code


def pipeline_serialize_file(
    input_file: Path,
    pipeline_info: str | None,
) -> tuple[Any, str, dict | None]:
    parser = Parser()
    yara_source = read_text(input_file)
    ast = parser.parse(yara_source)
    pipeline_data = json.loads(pipeline_info) if pipeline_info else None
    yaml_content = serialize_for_pipeline(ast, pipeline_data)
    return ast, yaml_content, pipeline_data


def build_rules_manifest(ast: YaraFile) -> str:
    return create_rules_manifest(ast)



"""Additional direct coverage for BaseVisitor mixins."""

from __future__ import annotations

from yaraast.ast.conditions import Condition
from yaraast.ast.expressions import BooleanLiteral, Expression
from yaraast.ast.strings import HexToken, StringDefinition
from yaraast.visitor.base import BaseVisitor


class _Visitor(BaseVisitor[None]):
    pass


def test_base_visitor_expression_and_condition_methods() -> None:
    visitor = _Visitor()
    condition = Condition()
    condition.expression = BooleanLiteral(value=True)

    assert visitor.visit_expression(Expression()) is None
    assert visitor.visit_condition(condition) is None


def test_base_visitor_string_definition_and_hex_token_methods() -> None:
    visitor = _Visitor()

    assert visitor.visit_string_definition(StringDefinition(identifier="$a")) is None
    assert visitor.visit_hex_token(HexToken()) is None



"""Property-based tests for YARA lexer using Hypothesis."""

from __future__ import annotations

import pytest
from hypothesis import assume, given, settings
from hypothesis import strategies as st

from yaraast.lexer.lexer import Lexer
from yaraast.lexer.lexer_tables import KEYWORDS as _LEXER_KEYWORDS
from yaraast.lexer.tokens import TokenType

_ALL_KEYWORDS = frozenset(_LEXER_KEYWORDS.keys())


def _yara_keyword() -> st.SearchStrategy[str]:
    """Generate YARA keywords."""
    return st.sampled_from(
        [
            "rule",
            "strings",
            "condition",
            "meta",
            "import",
            "include",
            "true",
            "false",
            "and",
            "or",
            "not",
            "all",
            "any",
            "of",
            "them",
            "for",
            "in",
            "at",
            "filesize",
            "entrypoint",
            "global",
            "private",
        ]
    )


def _valid_identifier() -> st.SearchStrategy[str]:
    return st.from_regex(r"[a-zA-Z_][a-zA-Z0-9_]{0,15}", fullmatch=True)


@pytest.mark.hypothesis
class TestLexerProperties:
    """Property-based tests for lexer behavior."""

    @given(rule_text=st.just("rule test { condition: true }"))
    @settings(max_examples=5, deadline=5000)
    def test_lexer_always_ends_with_eof(self, rule_text: str) -> None:
        """Every token stream ends with EOF."""
        lexer = Lexer(rule_text)
        tokens = lexer.tokenize()
        assert tokens[-1].type == TokenType.EOF

    @given(name=_valid_identifier())
    @settings(max_examples=50, deadline=5000)
    def test_simple_rule_tokenizes(self, name: str) -> None:
        """Simple rules always produce valid token streams."""
        assume(name.lower() not in _ALL_KEYWORDS)
        text = f"rule {name} {{ condition: true }}"
        lexer = Lexer(text)
        tokens = lexer.tokenize()
        assert len(tokens) >= 5  # rule, name, {, condition, :, true, }, EOF
        assert tokens[0].type == TokenType.RULE

    @given(text=_yara_keyword())
    @settings(max_examples=30, deadline=5000)
    def test_keywords_recognized(self, text: str) -> None:
        """All YARA keywords are recognized as keyword tokens (not identifiers)."""
        lexer = Lexer(text)
        tokens = lexer.tokenize()
        # Should have at least keyword + EOF
        assert len(tokens) >= 2
        # The keyword should not be tokenized as IDENTIFIER
        assert tokens[0].type != TokenType.IDENTIFIER or text not in Lexer.KEYWORDS

    @given(
        name=_valid_identifier(),
        n=st.integers(min_value=1, max_value=3),
    )
    @settings(max_examples=30, deadline=5000)
    def test_double_tokenize_idempotent(self, name: str, n: int) -> None:
        """Tokenizing the same text multiple times produces identical results."""
        assume(name.lower() not in _ALL_KEYWORDS)
        text = f"rule {name} {{ condition: true }}"
        tokens1 = Lexer(text).tokenize()
        tokens2 = Lexer(text).tokenize()
        assert len(tokens1) == len(tokens2)
        for t1, t2 in zip(tokens1, tokens2):
            assert t1.type == t2.type
            assert t1.value == t2.value



"""Diagnostic-code handlers for semantic LSP quick fixes."""

from __future__ import annotations

from lsprotocol.types import CodeAction, Diagnostic

from yaraast.lsp.code_action_semantic_quickfixes import (
    create_add_missing_arguments_action,
    create_add_placeholder_argument_action,
    create_import_module_action,
    create_replace_builtin_function_actions,
    create_replace_module_function_actions,
    create_trim_arguments_action,
)
from yaraast.lsp.lsp_docs import MODULE_DOCS


def handle_module_not_imported(
    _provider, _text: str, diagnostic: Diagnostic, uri: str, metadata
) -> list[CodeAction]:
    module_name = metadata.get("module")
    if isinstance(module_name, str):
        return create_import_module_action(module_name, diagnostic, uri)
    return []


def handle_module_function_not_found(
    provider, text: str, diagnostic: Diagnostic, uri: str, metadata
) -> list[CodeAction]:
    module_name = metadata.get("module")
    function_name = metadata.get("function")
    available = metadata.get("available_functions")
    if (
        isinstance(module_name, str)
        and isinstance(function_name, str)
        and isinstance(available, list)
    ):
        return create_replace_module_function_actions(
            text,
            diagnostic,
            uri,
            module_name,
            function_name,
            [item for item in available if isinstance(item, str)],
        )
    return []


def handle_invalid_arity(
    _provider, text: str, diagnostic: Diagnostic, uri: str, metadata
) -> list[CodeAction]:
    function_name = metadata.get("function")
    arity_kind = metadata.get("arity_kind")
    actual_args = metadata.get("actual_args")
    expected_min = metadata.get("expected_min")
    expected_args = metadata.get("expected_args")
    expected_max = metadata.get("expected_max")
    if (
        isinstance(function_name, str)
        and arity_kind == "min"
        and actual_args == 0
        and isinstance(expected_min, int)
        and expected_min >= 1
    ):
        return create_add_placeholder_argument_action(text, diagnostic, uri, function_name)
    if (
        isinstance(function_name, str)
        and arity_kind == "exact"
        and isinstance(actual_args, int)
        and isinstance(expected_args, int)
        and actual_args < expected_args
    ):
        return create_add_missing_arguments_action(
            text,
            diagnostic,
            uri,
            function_name,
            expected_args - actual_args,
        )
    if (
        isinstance(function_name, str)
        and arity_kind == "max"
        and isinstance(actual_args, int)
        and isinstance(expected_max, int)
        and actual_args > expected_max
    ):
        return create_trim_arguments_action(text, diagnostic, uri, function_name, expected_max)
    if (
        isinstance(function_name, str)
        and arity_kind == "exact"
        and isinstance(actual_args, int)
        and isinstance(expected_args, int)
        and actual_args > expected_args
    ):
        return create_trim_arguments_action(text, diagnostic, uri, function_name, expected_args)
    return []


def handle_unknown_function(
    _provider, text: str, diagnostic: Diagnostic, uri: str, metadata
) -> list[CodeAction]:
    function_name = metadata.get("function")
    suggested = metadata.get("suggested_functions")
    if isinstance(function_name, str) and isinstance(suggested, list):
        return create_replace_builtin_function_actions(
            text,
            diagnostic,
            uri,
            function_name,
            [item for item in suggested if isinstance(item, str)],
        )
    return []


def handle_duplicate_string_identifier(
    provider, text: str, diagnostic: Diagnostic, uri: str, metadata
) -> list[CodeAction]:
    identifier = metadata.get("identifier")
    if isinstance(identifier, str):
        return provider._create_rename_duplicate_action_from_identifier(
            text, diagnostic, uri, identifier
        )
    return []


def handle_validation_or_undefined(
    provider, text: str, diagnostic: Diagnostic, uri: str, metadata
) -> list[CodeAction]:
    identifier = metadata.get("identifier")
    if isinstance(identifier, str) and identifier.startswith("$"):
        return provider._create_add_string_action_from_identifier(text, diagnostic, uri, identifier)
    module_name = metadata.get("module")
    if isinstance(module_name, str) and module_name in MODULE_DOCS:
        return create_import_module_action(module_name, diagnostic, uri)
    return []



"""Text-based fallback helpers for LSP references and renames."""

from __future__ import annotations

from collections.abc import Iterable

from lsprotocol.types import Position

from yaraast.lsp.document_query_common import whole_word_positions
from yaraast.lsp.structure import SECTION_NAMES, get_rule_text_range


def iter_reference_occurrences(
    ctx, variants: Iterable[str], *, allowed_sections: tuple[str, ...]
) -> Iterable[tuple[int, int, str, str | None]]:
    for rule_start, rule_end in iter_rule_text_ranges(ctx):
        for line_num in range(rule_start, rule_end + 1):
            masked_line = mask_non_code_segments(ctx.lines[line_num])
            for variant in variants:
                for col in whole_word_positions(masked_line, variant):
                    section_name = section_for_occurrence(ctx.lines, rule_start, line_num, col)
                    if section_name not in allowed_sections:
                        continue
                    yield line_num, col, variant, section_name


def iter_rule_text_ranges(ctx) -> Iterable[tuple[int, int]]:
    seen: set[tuple[int, int]] = set()
    for symbol in ctx._symbols_of_kind("rule_block"):
        start = symbol.range.start.line
        end = symbol.range.end.line
        key = (start, end)
        if key in seen:
            continue
        seen.add(key)
        rule_text_range = get_rule_text_range(ctx.text, start)
        if rule_text_range is not None:
            yield rule_text_range.start, rule_text_range.end
        else:
            yield start, end


def section_for_occurrence(
    lines: list[str], rule_start: int, line_num: int, col: int
) -> str | None:
    current: str | None = None
    for idx in range(rule_start, line_num):
        stripped = lines[idx].strip()
        for section_name in SECTION_NAMES:
            if stripped == f"{section_name}:":
                current = section_name
                break
    line = lines[line_num]
    inline_sections: list[tuple[int, str]] = []
    for section_name in SECTION_NAMES:
        marker = f"{section_name}:"
        marker_idx = line.find(marker)
        if marker_idx >= 0 and marker_idx <= col:
            inline_sections.append((marker_idx, section_name))
    if inline_sections:
        inline_sections.sort(key=lambda item: item[0])
        current = inline_sections[-1][1]
    return current


def mask_non_code_segments(line: str) -> str:
    chars = list(line)
    in_string = False
    escape = False
    for idx, char in enumerate(chars):
        if escape:
            chars[idx] = " "
            escape = False
            continue
        if char == "\\" and in_string:
            chars[idx] = " "
            escape = True
            continue
        if char == '"':
            chars[idx] = " "
            in_string = not in_string
            continue
        if in_string:
            chars[idx] = " "
            continue
        if char == "/" and idx + 1 < len(chars) and chars[idx + 1] == "/":
            for comment_idx in range(idx, len(chars)):
                chars[comment_idx] = " "
            break
    return "".join(chars)


def line_has_assignment(line: str, end_idx: int) -> bool:
    return line[end_idx:].lstrip().startswith("=")


def matches_resolved_symbol(
    ctx, position: Position, *, kind: str | tuple[str, ...], normalized_name: str
) -> bool:
    resolved = ctx.resolve_symbol(position)
    valid_kinds = (kind,) if isinstance(kind, str) else kind
    return bool(
        resolved is not None
        and resolved.kind in valid_kinds
        and resolved.normalized_name == normalized_name
    )



"""Additional real coverage for simple_roundtrip_helpers."""

from __future__ import annotations

from pathlib import Path

from yaraast.ast.base import YaraFile
from yaraast.ast.expressions import BooleanLiteral
from yaraast.ast.meta import Meta
from yaraast.ast.rules import Import, Include, Rule, Tag
from yaraast.ast.strings import HexByte, HexString, PlainString, RegexString, StringDefinition
from yaraast.serialization.simple_roundtrip_helpers import (
    _compare_normalized,
    deserialize_from_file,
    deserialize_meta,
    deserialize_rule,
    deserialize_string,
    serialize_meta,
    serialize_rule,
    serialize_string,
    serialize_to_file,
    validate_roundtrip,
)


def test_simple_roundtrip_helpers_serialize_meta_and_string_fallbacks(tmp_path: Path) -> None:
    rule = Rule(
        name="helper_rule",
        condition=BooleanLiteral(value=True),
        tags=[Tag(name="one"), "two"],
        meta=[Meta(key="author", value="me"), Meta(key="enabled", value=True)],
        strings=[
            PlainString(identifier="$a", value="x"),
            HexString(identifier="$b", tokens=[HexByte(value=0x41)]),
            RegexString(identifier="$c", regex="ab.*"),
        ],
    )

    serialized_rule = serialize_rule(rule)
    assert serialized_rule["meta"][0] == {"type": "Meta", "key": "author", "value": "me"}
    assert serialize_meta(Meta(key="score", value=7)) == {
        "type": "Meta",
        "key": "score",
        "value": 7,
    }
    assert serialize_string(StringDefinition(identifier="$z"))["type"] == "StringDefinition"

    restored_rule = deserialize_rule(serialized_rule)
    assert [t.name for t in restored_rule.tags] == ["one", "two"]
    assert deserialize_meta({"key": "author", "value": "me"}).key == "author"
    assert deserialize_string({"type": "Unknown", "identifier": "$x", "data": "raw"}).value == "raw"

    path = tmp_path / "helper.json"
    serialize_to_file(
        YaraFile(imports=[Import(module="pe")], includes=[Include(path="inc.yar")], rules=[rule]),
        path,
    )
    restored_file = deserialize_from_file(path)
    assert isinstance(restored_file, YaraFile)
    assert restored_file.rules[0].name == "helper_rule"


def test_simple_roundtrip_helpers_compare_and_error_paths(tmp_path: Path) -> None:
    ok, differences = _compare_normalized("a\nb\nc", "a\nb\nc")
    assert ok is True
    assert differences == []

    ok_equal_len, differences_equal_len = _compare_normalized("a\nb\nc", "a\nx\nc")
    assert ok_equal_len is False
    assert differences_equal_len == ["Line 2 differs: 'b' vs 'x'"]

    ok2, differences2 = _compare_normalized(
        "1\n2\n3\n4\n5\n6\n7\n8",
        "x\ny\nz\nu\nv\nw",
    )
    assert ok2 is False
    assert differences2[0].startswith("Line count differs:")
    assert differences2[-1] == "... more differences"

    bad_json = tmp_path / "bad.json"
    bad_json.write_text("{not-json")
    try:
        deserialize_from_file(bad_json)
    except Exception:
        pass
    else:
        raise AssertionError("deserialize_from_file should fail on invalid JSON")

    valid, diff = validate_roundtrip(None)  # type: ignore[arg-type]
    assert valid is False
    assert "error" in diff

    fallback = deserialize_string({"type": "HexString", "identifier": "$h", "tokens": "{ 41 }"})
    assert isinstance(fallback, HexString)  # preserves type instead of converting to PlainString
    assert fallback.tokens == []

    default_condition_rule = deserialize_rule({"name": "fallback", "condition": None})
    assert isinstance(default_condition_rule.condition, BooleanLiteral)



"""More tests for performance optimizer (no mocks)."""

from __future__ import annotations

from pathlib import Path
from textwrap import dedent

from yaraast.parser import Parser
from yaraast.performance.optimizer import PerformanceOptimizer, optimize_yara_file


def _parse_yara(code: str):
    parser = Parser()
    return parser.parse(dedent(code))


def test_performance_optimizer_rule_and_file() -> None:
    code = """
    rule perf_opt {
        strings:
            $a = "abcd"
            $b = /ab+c/
            $c = { 6A 40 ?? }
        condition:
            $a or $b or $c
    }
    rule perf_opt2 {
        strings:
            $a = "a"
        condition:
            $a
    }
    """
    ast = _parse_yara(code)
    optimizer = PerformanceOptimizer()

    rule = ast.rules[0]
    optimized_rule = optimizer.optimize_rule(rule, strategy="balanced")
    assert optimized_rule is rule

    optimized_file = optimizer.optimize(ast, strategy="speed")
    assert optimized_file is ast

    stats = optimizer.get_statistics()
    assert stats["rules_optimized"] >= 1
    optimizer.reset_statistics()
    assert optimizer.get_statistics()["rules_optimized"] == 0


def test_optimize_yara_file(tmp_path: Path) -> None:
    code = """
    rule perf_opt_file {
        strings:
            $a = "abc"
        condition:
            $a
    }
    """
    path = tmp_path / "perf.yar"
    out = tmp_path / "perf_out.yar"
    path.write_text(dedent(code), encoding="utf-8")

    ast, stats = optimize_yara_file(str(path), output_path=str(out), strategy="memory")
    assert ast.rules
    assert out.exists()
    assert "rules_optimized" in stats



"""Outcome/match/options optimization mixin for YARA-L optimizer."""

from __future__ import annotations

from typing import Any

from yaraast.yaral.ast_nodes import MatchSection, OutcomeAssignment, OutcomeSection


class YaraLOptimizerOutcomeMixin:
    """Outcome, match, and options optimization methods."""

    def _optimize_match_section(self, match: MatchSection) -> MatchSection:
        if not match:
            return match

        optimized_vars = []
        for var in match.variables:
            var_name = getattr(var, "variable", getattr(var, "name", None))
            if var_name and self._is_match_var_used(var_name):
                if hasattr(var, "time_window") and var.time_window:
                    var.time_window = self._optimize_time_window(var.time_window)
                optimized_vars.append(var)

        return MatchSection(variables=optimized_vars)

    def _optimize_time_window(self, window: Any) -> Any:
        if hasattr(window, "duration") and hasattr(window, "unit"):
            duration = window.duration
            unit = window.unit

            if unit in ["s", "seconds"] and duration >= 3600:
                window.duration = duration // 3600
                window.unit = "h"
                self.stats.time_windows_optimized += 1
            elif unit in ["m", "minutes"] and duration >= 1440:
                window.duration = duration // 1440
                window.unit = "d"
                self.stats.time_windows_optimized += 1

        return window

    def _optimize_outcome_section(self, outcome: OutcomeSection) -> OutcomeSection:
        if not outcome:
            return outcome

        if hasattr(outcome, "assignments"):
            return outcome

        optimized_vars = {}
        if hasattr(outcome, "variables"):
            for var_name, var_expr in outcome.variables.items():
                if self._is_outcome_var_used(var_name) or var_name in [
                    "risk_score",
                    "severity",
                ]:
                    optimized_vars[var_name] = var_expr

        return OutcomeSection(
            assignments=[
                OutcomeAssignment(variable=var_name, expression=var_expr)
                for var_name, var_expr in optimized_vars.items()
            ]
        )

    def _optimize_options(self, options: Any) -> Any:
        if not options:
            return options

        if hasattr(options, "options"):
            if "max_events" not in options.options:
                options.options["max_events"] = 10000

            if "timeout" not in options.options:
                options.options["timeout"] = "5m"

        return options



"""Additional branch coverage for YARA-X collection parser helpers (no mocks)."""

from __future__ import annotations

from yaraast.ast.expressions import Identifier
from yaraast.lexer.tokens import Token, TokenType
from yaraast.yarax.ast_nodes import (
    ArrayComprehension,
    DictComprehension,
    DictExpression,
    ListExpression,
    SliceExpression,
    SpreadOperator,
    TupleExpression,
    TupleIndexing,
)
from yaraast.yarax.parser import YaraXParser


def _tok(token_type: TokenType, value: str | int | None) -> Token:
    return Token(token_type, value, 1, 1)


def _manual_parser(tokens: list[Token]) -> YaraXParser:
    parser = YaraXParser("rule seed { condition: true }")
    parser.tokens = [*tokens, _tok(TokenType.EOF, None)]
    parser.current = 0
    return parser


def _parse_expr(text: str):
    return YaraXParser(text).parse_expression()


def test_empty_and_trailing_collection_variants() -> None:
    assert _parse_expr("[]") == ListExpression(elements=[])

    empty_dict = _parse_expr("{}")
    assert isinstance(empty_dict, DictExpression)
    assert empty_dict.items == []

    empty_tuple = _parse_expr("()")
    assert isinstance(empty_tuple, TupleExpression)
    assert empty_tuple.elements == []

    trailing = _parse_expr('[1, "x",]')
    assert isinstance(trailing, ListExpression)
    assert len(trailing.elements) == 2

    trailing_dict = _parse_expr('{"a": 1,}')
    assert isinstance(trailing_dict, DictExpression)
    assert len(trailing_dict.items) == 1


def test_spread_and_keyword_keyword_fallback_paths() -> None:
    parser = _manual_parser(
        [
            _tok(TokenType.LBRACKET, "["),
            _tok(TokenType.DOUBLE_DOT, ".."),
            _tok(TokenType.DOT, "."),
            _tok(TokenType.IDENTIFIER, "arr"),
            _tok(TokenType.RBRACKET, "]"),
        ],
    )
    spread_list = parser._parse_list_or_comprehension()
    assert isinstance(spread_list, ListExpression)
    assert isinstance(spread_list.elements[0], SpreadOperator)

    parser = _manual_parser(
        [
            _tok(TokenType.LBRACKET, "["),
            _tok(TokenType.IDENTIFIER, "x"),
            _tok(TokenType.IDENTIFIER, "for"),
            _tok(TokenType.IDENTIFIER, "x"),
            _tok(TokenType.IDENTIFIER, "in"),
            _tok(TokenType.IDENTIFIER, "items"),
            _tok(TokenType.RBRACKET, "]"),
        ],
    )
    comp = parser._parse_list_or_comprehension()
    assert isinstance(comp, ArrayComprehension)
    assert comp.variable == "x"

    parser = _manual_parser(
        [
            _tok(TokenType.LBRACE, "{"),
            _tok(TokenType.IDENTIFIER, "k"),
            _tok(TokenType.COLON, ":"),
            _tok(TokenType.IDENTIFIER, "v"),
            _tok(TokenType.IDENTIFIER, "for"),
            _tok(TokenType.IDENTIFIER, "k"),
            _tok(TokenType.COMMA, ","),
            _tok(TokenType.IDENTIFIER, "v"),
            _tok(TokenType.IDENTIFIER, "in"),
            _tok(TokenType.IDENTIFIER, "data"),
            _tok(TokenType.IDENTIFIER, "if"),
            _tok(TokenType.IDENTIFIER, "v"),
            _tok(TokenType.RBRACE, "}"),
        ],
    )
    dict_comp = parser._parse_dict_or_comprehension()
    assert isinstance(dict_comp, DictComprehension)
    assert dict_comp.value_variable == "v"
    assert dict_comp.condition is not None


def test_tuple_parentheses_and_slice_paths() -> None:
    parser = _manual_parser(
        [
            _tok(TokenType.LPAREN, "("),
            _tok(TokenType.IDENTIFIER, "x"),
            _tok(TokenType.RPAREN, ")"),
            _tok(TokenType.LBRACKET, "["),
        ],
    )
    single_tuple = parser._parse_tuple_or_parentheses()
    assert isinstance(single_tuple, TupleExpression)
    assert len(single_tuple.elements) == 1

    parser = _manual_parser(
        [
            _tok(TokenType.LBRACKET, "["),
            _tok(TokenType.COLON, ":"),
            _tok(TokenType.INTEGER, 3),
            _tok(TokenType.RBRACKET, "]"),
        ],
    )
    idx = parser._parse_tuple_indexing_postfix(Identifier(name="x"))
    assert isinstance(idx, SliceExpression)
    assert idx.start is None
    assert idx.stop is not None

    parser = _manual_parser(
        [
            _tok(TokenType.LBRACKET, "["),
            _tok(TokenType.INTEGER, 1),
            _tok(TokenType.RBRACKET, "]"),
        ],
    )
    idx = parser._parse_tuple_indexing_postfix(Identifier(name="x"))
    assert isinstance(idx, TupleIndexing)

    parser = _manual_parser(
        [
            _tok(TokenType.LBRACKET, "["),
            _tok(TokenType.INTEGER, 1),
            _tok(TokenType.COLON, ":"),
            _tok(TokenType.INTEGER, 4),
            _tok(TokenType.COLON, ":"),
            _tok(TokenType.INTEGER, 2),
            _tok(TokenType.RBRACKET, "]"),
        ],
    )
    idx = parser._parse_tuple_indexing_postfix(Identifier(name="x"))
    assert isinstance(idx, SliceExpression)
    assert idx.start is not None
    assert idx.step is not None



"""Token protocol abstractions for parser and lexer contracts.

Note: IToken.type is typed as Enum to avoid depending on the concrete
lexer module.  The canonical implementation uses yaraast.lexer.tokens.TokenType.
"""

from __future__ import annotations

from enum import Enum
from typing import Any, Protocol


class IToken(Protocol):
    """Minimal token contract consumed by the parser."""

    type: Enum
    value: Any
    line: int
    column: int
    length: int



"""Additional real coverage for AST transformer and fluent rule builder."""

from __future__ import annotations

from yaraast.ast.base import YaraFile
from yaraast.ast.expressions import BooleanLiteral, StringIdentifier
from yaraast.ast.rules import Rule, Tag
from yaraast.ast.strings import PlainString, RegexString
from yaraast.builder.ast_transformer import (
    RuleTransformer,
    YaraFileTransformer,
    create_variant_rule,
)
from yaraast.builder.fluent_condition_builder import FluentConditionBuilder
from yaraast.builder.fluent_rule_builder import FluentRuleBuilder


def test_fluent_rule_builder_additional_builder_paths() -> None:
    built = (
        FluentRuleBuilder("extra")
        .with_string("$plain", "alpha")
        .pe_header("$pe")
        .email_pattern("$mail")
        .string("$re")
        .regex("foo.*bar")
        .then()
        .condition(FluentConditionBuilder().true())
        .for_small_files()
        .build()
    )

    assert {s.identifier for s in built.strings} == {"$plain", "$pe", "$mail", "$re"}
    assert isinstance(next(s for s in built.strings if s.identifier == "$re"), RegexString)
    assert built.condition is not None
    assert isinstance(built.meta, list)

    built_with_fluent_condition = (
        FluentRuleBuilder("fcb")
        .condition(
            FluentConditionBuilder().any_of_them(),
        )
        .build()
    )
    assert built_with_fluent_condition.condition is not None


def test_ast_transformer_noop_and_list_meta_paths() -> None:
    list_meta_rule = Rule(
        name="list_meta",
        tags=[Tag(name="tag1"), Tag(name="tag2")],
        meta=[],
        strings=[PlainString(identifier="$a", value="x")],
        condition=StringIdentifier(name="$a"),
    )

    transformed = (
        RuleTransformer(list_meta_rule)
        .add_tag("tag1")
        .add_meta("author", "ignored")
        .remove_meta("author")
        .remove_string("missing")
        .transform_condition(lambda expr: BooleanLiteral(value=False))
        .build()
    )

    assert [t.name for t in transformed.tags] == ["tag1", "tag2"]
    assert transformed.meta == []
    assert len(transformed.strings) == 1
    assert isinstance(transformed.condition, BooleanLiteral)
    assert transformed.condition.value is False

    dict_meta_rule = Rule(
        name="dict_meta",
        meta={"author": "me", "version": 1},
        strings=[
            PlainString(identifier="$a", value="x"),
            PlainString(identifier="$b", value="y"),
        ],
        condition=StringIdentifier(name="$b"),
    )
    transformed_dict = (
        RuleTransformer(dict_meta_rule)
        .remove_meta("author")
        .rename_strings({"$a": "$renamed"})
        .build()
    )
    assert len(transformed_dict.meta) == 1
    assert transformed_dict.get_meta_value("version") == 1
    assert [s.identifier for s in transformed_dict.strings] == ["$renamed", "$b"]
    assert isinstance(transformed_dict.condition, StringIdentifier)
    assert transformed_dict.condition.name == "$b"

    no_condition = Rule(name="empty")
    untouched = (
        RuleTransformer(no_condition)
        .transform_condition(lambda expr: BooleanLiteral(value=True))
        .build()
    )
    assert untouched.condition is None


def test_yara_file_transformer_no_match_and_variant_without_optional_changes() -> None:
    rule_one = Rule(name="one")
    rule_two = Rule(name="two")
    yara_file = YaraFile(rules=[rule_one, rule_two])

    unchanged = (
        YaraFileTransformer(yara_file)
        .transform_rule("missing", lambda r: RuleTransformer(r).rename("nope").build())
        .build()
    )
    assert [r.name for r in unchanged.rules] == ["one", "two"]

    variant = create_variant_rule(rule_one, "variant_only")
    assert variant.name == "variant_only"
    assert variant.tags == []



"""Document highlight provider for YARAAST LSP."""

from __future__ import annotations

from lsprotocol.types import DocumentHighlight, DocumentHighlightKind, Position

from yaraast.lsp.document_context import DocumentContext
from yaraast.lsp.document_highlight_helpers import (
    highlight_identifier,
    highlight_string_identifier,
    simple_highlight,
)
from yaraast.lsp.document_types import ReferenceRecord
from yaraast.lsp.utils import get_word_at_position


class DocumentHighlightProvider:
    """Provide document highlighting for symbols."""

    def get_highlights(self, text: str, position: Position) -> list[DocumentHighlight]:
        """Get all highlights for symbol at position."""
        word, _word_range = get_word_at_position(text, position)
        if not word:
            return []

        # Check if it's a string identifier
        if word.startswith("$"):
            return self._highlight_string_identifier(text, word)

        # Check if it's a rule name or other identifier
        return self._highlight_identifier(text, word)

    def _highlight_string_identifier(self, text: str, identifier: str) -> list[DocumentHighlight]:
        """Highlight all occurrences of a string identifier."""
        records = self._get_string_reference_records(text, identifier)
        if records:
            return self._highlights_from_records(records)
        return highlight_string_identifier(text, identifier)

    def _highlight_identifier(self, text: str, identifier: str) -> list[DocumentHighlight]:
        """Highlight all occurrences of a regular identifier."""
        records = self._get_rule_reference_records(text, identifier)
        if records:
            return self._highlights_from_records(records)
        return highlight_identifier(text, identifier)

    def _simple_highlight(self, text: str, word: str) -> list[DocumentHighlight]:
        """Simple text-based highlighting fallback."""
        return simple_highlight(text, word)

    def _get_string_reference_records(self, text: str, identifier: str) -> list[ReferenceRecord]:
        ctx = DocumentContext("file:///document-highlight.yar", text)
        if ctx.ast() is None:
            return []
        return ctx.find_string_reference_records(identifier, include_declaration=True)

    def _get_rule_reference_records(self, text: str, identifier: str) -> list[ReferenceRecord]:
        ctx = DocumentContext("file:///document-highlight.yar", text)
        if ctx.ast() is None:
            return []
        return ctx.rule_reference_records(identifier, include_declaration=True)

    def _highlights_from_records(self, records: list[ReferenceRecord]) -> list[DocumentHighlight]:
        highlights: list[DocumentHighlight] = []
        for record in records:
            kind = (
                DocumentHighlightKind.Write
                if record.role == "declaration"
                else DocumentHighlightKind.Read
            )
            highlights.append(DocumentHighlight(range=record.location.range, kind=kind))
        return highlights



"""CLI command for performance benchmarks of AST operations."""

from __future__ import annotations

from pathlib import Path

import click
from rich.console import Console

from yaraast.cli.bench_reporting import (
    display_benchmark_file,
    display_benchmark_summary,
    display_operation_result,
    display_performance_comparison,
    print_benchmark_header,
    save_benchmark_results,
)
from yaraast.cli.bench_services import (
    _determine_operations_to_run,
    _get_benchmark_summary,
    _run_single_operation,
)

console = Console()


@click.command()
@click.argument("files", nargs=-1, type=click.Path(exists=True), required=True)
@click.option(
    "--operations",
    type=click.Choice(["parse", "codegen", "roundtrip", "all"]),
    default="all",
    help="Operations to benchmark",
)
@click.option(
    "--iterations",
    type=int,
    default=10,
    help="Number of iterations per test",
)
@click.option(
    "--output",
    type=click.Path(),
    help="Output benchmark results to JSON file",
)
@click.option("--compare", is_flag=True, help="Compare performance across files")
def bench(
    files: tuple[str],
    operations: str,
    iterations: int,
    output: str | None,
    compare: bool,
) -> None:
    """Performance benchmarks for AST operations."""
    try:
        from yaraast.cli.ast_tools import ASTBenchmarker

        file_paths = [Path(f) for f in files]
        benchmarker = ASTBenchmarker()

        print_benchmark_header(file_paths, iterations)
        all_results = []
        ops_to_run = _determine_operations_to_run(operations)
        for file_path in file_paths:
            display_benchmark_file(file_path)
            file_results = {}
            for op in ops_to_run:
                result = _run_single_operation(benchmarker, file_path, op, iterations)
                if result:
                    display_operation_result(op, result)
                    if result.success:
                        file_results[op] = result

            all_results.append(
                {
                    "file": str(file_path),
                    "file_name": file_path.name,
                    "results": file_results,
                }
            )
        summary = _get_benchmark_summary(benchmarker)
        display_benchmark_summary(summary)

        if compare and len(file_paths) > 1:
            display_performance_comparison(all_results)

        if output:
            save_benchmark_results(output, iterations, operations, all_results, summary)

        console.print("\nBenchmarking completed!")

    except Exception as e:
        from rich.markup import escape

        console.print(f"[red]Error: {escape(str(e))}[/red]")
        raise click.Abort from None



"""More tests for evaluator (no mocks)."""

from __future__ import annotations

from yaraast.ast.base import YaraFile
from yaraast.ast.conditions import OfExpression
from yaraast.ast.expressions import (
    BinaryExpression,
    BooleanLiteral,
    FunctionCall,
    Identifier,
    IntegerLiteral,
    StringIdentifier,
    StringLiteral,
    UnaryExpression,
)
from yaraast.ast.modules import ModuleReference
from yaraast.ast.operators import DefinedExpression
from yaraast.ast.rules import Import, Rule
from yaraast.ast.strings import PlainString
from yaraast.evaluation.evaluator import YaraEvaluator


def test_evaluator_function_calls_and_math_module() -> None:
    data = b"\x01\x02\x03\x04"
    evaluator = YaraEvaluator(data=data)
    ast = YaraFile(
        imports=[Import(module="math")],
        rules=[
            Rule(
                name="r1",
                condition=BinaryExpression(
                    left=FunctionCall(function="math.abs", arguments=[IntegerLiteral(value=-5)]),
                    operator="==",
                    right=IntegerLiteral(value=5),
                ),
            ),
        ],
    )

    results = evaluator.evaluate_file(ast)
    assert results["r1"] is True


def test_evaluator_string_ops_and_defined() -> None:
    evaluator = YaraEvaluator(data=b"abcd")
    rule = Rule(
        name="r1",
        strings=[PlainString(identifier="$a", value="ab")],
        condition=BinaryExpression(
            left=StringLiteral(value="Hello"),
            operator="icontains",
            right=StringLiteral(value="he"),
        ),
    )
    assert evaluator.evaluate_rule(rule) is True

    rule.condition = DefinedExpression(expression=StringIdentifier(name="$a"))
    assert evaluator.evaluate_rule(rule) is True


def test_evaluator_unary_and_of_expression() -> None:
    evaluator = YaraEvaluator(data=b"aaaa")
    rule = Rule(
        name="r1",
        strings=[
            PlainString(identifier="$a", value="aa"),
            PlainString(identifier="$b", value="bb"),
        ],
        condition=UnaryExpression(operator="not", operand=BooleanLiteral(value=False)),
    )
    assert evaluator.evaluate_rule(rule) is True

    rule.condition = OfExpression(quantifier="any", string_set=Identifier(name="them"))
    assert evaluator.evaluate_rule(rule) is True


def test_evaluator_module_reference() -> None:
    evaluator = YaraEvaluator(data=b"MZ")
    ast = YaraFile(imports=[Import(module="pe")], rules=[Rule(name="r1")])
    evaluator.evaluate_file(ast)
    module = evaluator.visit_module_reference(ModuleReference(module="pe"))
    assert module is not None



"""Parsing routines for YARA-L rules."""

from __future__ import annotations

from yaraast.lexer.tokens import TokenType as BaseTokenType

from ._shared import YaraLParserError
from .ast_nodes import (
    BinaryCondition,
    ConditionExpression,
    ConditionSection,
    EventCountCondition,
    EventExistsCondition,
    UnaryCondition,
    VariableComparisonCondition,
)
from .tokens import YaraLTokenType


class YaraLConditionParsingMixin:
    """Mixin providing YARA-L parse routines."""

    def _parse_condition_section(self) -> ConditionSection:
        """Parse condition section."""
        self._consume_keyword("condition")
        self._consume(BaseTokenType.COLON, "Expected ':' after 'condition'")

        expression = self._parse_condition_expression()

        return ConditionSection(expression=expression)

    def _parse_condition_expression(self) -> ConditionExpression:
        """Parse condition expression."""
        return self._parse_or_condition()

    def _parse_or_condition(self) -> ConditionExpression:
        """Parse OR condition."""
        left = self._parse_and_condition()

        while self._check_keyword("or"):
            self._advance()
            right = self._parse_and_condition()
            left = BinaryCondition(operator="or", left=left, right=right)

        return left

    def _parse_and_condition(self) -> ConditionExpression:
        """Parse AND condition."""
        left = self._parse_unary_condition()

        while self._check_keyword("and"):
            self._advance()
            right = self._parse_unary_condition()
            left = BinaryCondition(operator="and", left=left, right=right)

        return left

    def _parse_unary_condition(self) -> ConditionExpression:
        """Parse unary condition."""
        if self._check_keyword("not"):
            self._advance()
            operand = self._parse_unary_condition()
            return UnaryCondition(operator="not", operand=operand)

        return self._parse_primary_condition()

    def _parse_primary_condition(self) -> ConditionExpression:
        """Parse primary condition."""
        # Parenthesized expression
        if self._check(BaseTokenType.LPAREN):
            return self._parse_parenthesized_condition()

        # Event count: #e > 5
        if self._check(BaseTokenType.STRING_COUNT):
            return self._parse_event_count_condition()

        # Variable or event reference: $var or $e1
        if self._check_yaral_type(YaraLTokenType.EVENT_VAR) or self._check(
            BaseTokenType.STRING_IDENTIFIER
        ):
            return self._parse_variable_condition()

        # Fallback: treat as exists condition
        if self._check(BaseTokenType.IDENTIFIER):
            return self._parse_identifier_condition()

        msg = f"Unexpected token in condition: {self._peek()}"
        raise YaraLParserError(
            msg,
            self._peek(),
        )

    def _parse_parenthesized_condition(self) -> ConditionExpression:
        """Parse a parenthesized condition expression."""
        self._advance()
        expr = self._parse_condition_expression()
        self._consume(BaseTokenType.RPAREN, "Expected ')' after expression")
        return expr

    def _parse_event_count_condition(self) -> EventCountCondition:
        """Parse an event count condition: #e > 5."""
        self._advance()
        event_name = self._consume(
            BaseTokenType.IDENTIFIER,
            "Expected event name after '#'",
        ).value

        operator = self._consume_comparison_operator()
        count = int(
            self._consume(
                BaseTokenType.INTEGER,
                "Expected number after operator",
            ).value,
        )

        return EventCountCondition(event=event_name, operator=operator, count=count)

    def _consume_comparison_operator(self) -> str:
        """Consume and return a comparison operator token."""
        operator_map = {
            BaseTokenType.GT: ">",
            BaseTokenType.LT: "<",
            BaseTokenType.GE: ">=",
            BaseTokenType.LE: "<=",
            BaseTokenType.EQ: "==",
            BaseTokenType.NEQ: "!=",
        }
        for token_type, op in operator_map.items():
            if self._check(token_type):
                self._advance()
                return op

        msg = "Expected comparison operator"
        raise YaraLParserError(msg, self._peek())

    def _parse_comparison_value(self):
        """Parse the value on the right side of a comparison."""
        if self._check(BaseTokenType.INTEGER):
            return int(self._advance().value)
        if (
            self._check(BaseTokenType.STRING)
            or self._check_yaral_type(YaraLTokenType.EVENT_VAR)
            or self._check(BaseTokenType.STRING_IDENTIFIER)
            or self._check(BaseTokenType.IDENTIFIER)
        ):
            return self._advance().value

        msg = "Expected value after comparison operator"
        raise YaraLParserError(msg, self._peek())

    def _check_comparison_operator(self) -> bool:
        """Check if the current token is a comparison operator."""
        return (
            self._check(BaseTokenType.GT)
            or self._check(BaseTokenType.LT)
            or self._check(BaseTokenType.GE)
            or self._check(BaseTokenType.LE)
            or self._check(BaseTokenType.EQ)
            or self._check(BaseTokenType.NEQ)
        )

    def _parse_variable_condition(self) -> ConditionExpression:
        """Parse a condition starting with a variable ($var or $e1)."""
        var_token = self._advance()
        var_name = var_token.value

        # Check if followed by comparison operator
        if self._check_comparison_operator():
            operator = self._consume_comparison_operator()
            value = self._parse_comparison_value()
            return VariableComparisonCondition(variable=var_name, operator=operator, value=value)

        # Just a variable reference (event exists)
        event_name = var_name.lstrip("$")
        return EventExistsCondition(event=event_name)

    def _parse_identifier_condition(self) -> ConditionExpression:
        """Parse a condition starting with an identifier."""
        name = self._advance().value

        # Check if followed by comparison operator
        if self._check_comparison_operator():
            operator = self._consume_comparison_operator()
            value = self._parse_comparison_value()
            return VariableComparisonCondition(variable=name, operator=operator, value=value)

        return EventExistsCondition(event=name)



"""Dependency analyzer for YARA rules."""

from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any

from yaraast.visitor.base import BaseVisitor

if TYPE_CHECKING:
    from yaraast.ast.base import YaraFile
    from yaraast.ast.expressions import FunctionCall, Identifier
    from yaraast.ast.rules import Import, Include, Rule


class DependencyAnalyzer(BaseVisitor[None]):
    """Analyze dependencies between YARA rules."""

    def __init__(self) -> None:
        self.rule_names: set[str] = set()
        self.dependencies: dict[str, set[str]] = defaultdict(
            set,
        )  # rule -> rules it depends on
        self.current_rule: str | None = None
        self.imported_modules: set[str] = set()
        self.included_files: set[str] = set()

    def analyze(self, yara_file: YaraFile) -> dict[str, Any]:
        """Analyze dependencies in YARA file."""
        self.rule_names.clear()
        self.dependencies.clear()
        self.imported_modules.clear()
        self.included_files.clear()
        self.current_rule = None

        # First pass: collect all rule names
        for rule in yara_file.rules:
            self.rule_names.add(rule.name)

        # Visit imports and includes
        for imp in yara_file.imports:
            self.visit(imp)

        for inc in yara_file.includes:
            self.visit(inc)

        # Second pass: analyze dependencies
        for rule in yara_file.rules:
            self.visit(rule)

        return {
            "rules": list(self.rule_names),
            "dependencies": {rule: list(deps) for rule, deps in self.dependencies.items()},
            "dependency_graph": self._build_dependency_graph(),
            "circular_dependencies": self._find_circular_dependencies(),
            "dependency_order": self._topological_sort(),
            "imported_modules": list(self.imported_modules),
            "included_files": list(self.included_files),
        }

    def get_dependencies(self, rule_name: str) -> list[str]:
        """Get direct dependencies of a rule."""
        return list(self.dependencies.get(rule_name, set()))

    def get_dependents(self, rule_name: str) -> list[str]:
        """Get rules that depend on the given rule."""
        dependents = []
        for rule, deps in self.dependencies.items():
            if rule_name in deps:
                dependents.append(rule)
        return dependents

    def get_transitive_dependencies(self, rule_name: str) -> set[str]:
        """Get all transitive dependencies of a rule."""
        visited = set()
        to_visit = [rule_name]

        while to_visit:
            current = to_visit.pop()
            if current in visited:
                continue

            visited.add(current)
            deps = self.dependencies.get(current, set())
            to_visit.extend(deps - visited)

        visited.remove(rule_name)  # Don't include self
        return visited

    def _build_dependency_graph(self) -> dict[str, dict[str, list[str]]]:
        """Build a dependency graph."""
        graph = {}

        for rule in self.rule_names:
            deps = self.dependencies.get(rule, set())
            dependents = self.get_dependents(rule)
            transitive = self.get_transitive_dependencies(rule)

            graph[rule] = {
                "depends_on": list(deps),
                "depended_by": dependents,
                "transitive_dependencies": list(transitive),
                "is_independent": len(deps) == 0 and len(dependents) == 0,
            }

        return graph

    def _find_circular_dependencies(self) -> list[list[str]]:
        """Find circular dependencies using DFS."""
        dfs_state = self._init_dfs_state()
        cycles = []

        for rule in self.rule_names:
            if dfs_state["color"][rule] == dfs_state["white"]:
                self._dfs_cycle_detection(rule, dfs_state, cycles)

        return self._remove_duplicate_cycles(cycles)

    def _init_dfs_state(self) -> dict[str, Any]:
        """Initialize DFS state for cycle detection."""
        white, gray, black = 0, 1, 2
        return {
            "white": white,
            "gray": gray,
            "black": black,
            "color": dict.fromkeys(self.rule_names, white),
            "path": [],
        }

    def _dfs_cycle_detection(
        self, node: str, state: dict[str, Any], cycles: list[list[str]]
    ) -> None:
        """Perform DFS for cycle detection."""
        state["color"][node] = state["gray"]
        state["path"].append(node)

        for neighbor in self.dependencies.get(node, set()):
            if neighbor in self.rule_names:  # Only check internal rules
                if state["color"][neighbor] == state["gray"]:
                    # Found cycle - include the back-edge to close it
                    cycle_start = state["path"].index(neighbor)
                    cycles.append([*state["path"][cycle_start:], neighbor])
                elif state["color"][neighbor] == state["white"]:
                    self._dfs_cycle_detection(neighbor, state, cycles)

        state["path"].pop()
        state["color"][node] = state["black"]

    def _remove_duplicate_cycles(self, cycles: list[list[str]]) -> list[list[str]]:
        """Remove duplicate cycles from the list."""
        unique_cycles = []
        for cycle in cycles:
            # Normalize by rotating the cycle body (excluding closing back-edge)
            body = cycle[:-1] if len(cycle) > 1 and cycle[0] == cycle[-1] else cycle
            if body:
                normalized_body = min(body[i:] + body[:i] for i in range(len(body)))
                normalized = [*normalized_body, normalized_body[0]]  # Re-add closing edge
            else:
                normalized = cycle
            if normalized not in unique_cycles:
                unique_cycles.append(normalized)

        return unique_cycles

    def _topological_sort(self) -> list[str] | None:
        """Perform topological sort on rules."""
        # Check for cycles first
        if self._find_circular_dependencies():
            return None

        in_degree = dict.fromkeys(self.rule_names, 0)

        # Calculate in-degrees: if rule A depends on rule B, then A has incoming edge from B
        for rule, deps in self.dependencies.items():
            in_degree[rule] = len([dep for dep in deps if dep in in_degree])

        # Find nodes with no incoming edges
        queue = [rule for rule, degree in in_degree.items() if degree == 0]
        result = []

        while queue:
            current = queue.pop(0)
            result.append(current)

            # Reduce in-degree for dependent nodes
            for rule, deps in self.dependencies.items():
                if current in deps and rule in in_degree:
                    in_degree[rule] -= 1
                    if in_degree[rule] == 0:
                        queue.append(rule)

        return result if len(result) == len(self.rule_names) else None

    # Visitor methods with actual logic
    def visit_import(self, node: Import) -> None:
        self.imported_modules.add(node.module)

    def visit_include(self, node: Include) -> None:
        self.included_files.add(node.path)

    def visit_rule(self, node: Rule) -> None:
        self.current_rule = node.name

        # Check condition for rule references
        if node.condition:
            self.visit(node.condition)

        self.current_rule = None

    def visit_identifier(self, node: Identifier) -> None:
        # Check if identifier is a rule reference
        if self.current_rule and node.name in self.rule_names:
            self.dependencies[self.current_rule].add(node.name)

    def visit_function_call(self, node: FunctionCall) -> None:
        # Some functions might reference rules
        if node.function in self.rule_names and self.current_rule:
            self.dependencies[self.current_rule].add(node.function)

        # Visit arguments (BaseVisitor.visit_function_call does this, but we call super)
        super().visit_function_call(node)



from __future__ import annotations

from pathlib import Path

from lsprotocol.types import Position

from yaraast.lsp.definition import DefinitionProvider
from yaraast.lsp.hover import HoverProvider
from yaraast.lsp.references import ReferencesProvider
from yaraast.lsp.rename import RenameProvider
from yaraast.lsp.runtime import LspRuntime, path_to_uri
from yaraast.lsp.symbols import SymbolsProvider

FIXTURES = Path(__file__).parent / "fixtures" / "lsp_parity"


def _pos(line: int, char: int) -> Position:
    return Position(line=line, character=char)


def _runtime_for(folder: Path, dialect: str) -> tuple[LspRuntime, str, str]:
    runtime = LspRuntime()
    runtime.update_config({"YARA": {"dialectMode": dialect}})
    runtime.set_workspace_folders([str(folder)])
    user = folder / "user.yar"
    text = user.read_text(encoding="utf-8")
    uri = path_to_uri(user)
    runtime.open_document(uri, text)
    return runtime, text, uri


def test_lsp_parity_classic_requests() -> None:
    runtime, text, uri = _runtime_for(FIXTURES / "classic", "yara")
    hover = HoverProvider(runtime).get_hover(text, _pos(4, 10), uri)
    assert hover is not None and "**shared_rule**" in hover.contents.value

    definition = DefinitionProvider(runtime).get_definition(text, _pos(4, 10), uri)
    assert definition is not None
    assert definition.uri.endswith("/common.yar")

    refs = ReferencesProvider(runtime).get_references(text, _pos(4, 10), uri)
    assert len(refs) >= 2

    rename = RenameProvider(runtime).rename(text, _pos(4, 10), "renamed_rule", uri)
    assert rename is not None and rename.changes is not None
    assert len(rename.changes) == 2

    symbols = SymbolsProvider(runtime).get_symbols(text, uri)
    assert any(symbol.name == "wrapper" for symbol in symbols)


def test_lsp_parity_yaral_requests() -> None:
    runtime, text, uri = _runtime_for(FIXTURES / "yaral", "yaral")
    hover = HoverProvider(runtime).get_hover(text, _pos(2, 10), uri)
    assert hover is not None and "**detect_login**" in hover.contents.value

    definition = DefinitionProvider(runtime).get_definition(text, _pos(2, 10), uri)
    assert definition is not None
    assert definition.uri.endswith("/common.yar")

    refs = ReferencesProvider(runtime).get_references(text, _pos(2, 10), uri)
    assert len(refs) >= 2

    rename = RenameProvider(runtime).rename(text, _pos(2, 10), "detect_login_new", uri)
    assert rename is not None and rename.changes is not None
    assert len(rename.changes) == 2

    symbols = SymbolsProvider(runtime).get_symbols(text, uri)
    child_names = {child.name for child in (symbols[0].children or [])}
    assert {"condition"} <= child_names


def test_lsp_parity_yarax_requests() -> None:
    runtime, text, uri = _runtime_for(FIXTURES / "yarax", "yarax")
    hover = HoverProvider(runtime).get_hover(text, _pos(3, 8), uri)
    assert hover is not None and "**helper**" in hover.contents.value

    definition = DefinitionProvider(runtime).get_definition(text, _pos(3, 8), uri)
    assert definition is not None
    assert definition.uri.endswith("/common.yar")

    refs = ReferencesProvider(runtime).get_references(text, _pos(3, 8), uri)
    assert len(refs) >= 2

    rename = RenameProvider(runtime).rename(text, _pos(3, 8), "helper_new", uri)
    assert rename is not None and rename.changes is not None
    assert len(rename.changes) == 2

    symbols = SymbolsProvider(runtime).get_symbols(text, uri)
    assert symbols and symbols[0].name == "sample"



"""Simple differ for YARA files."""

from __future__ import annotations

from collections import Counter
from dataclasses import dataclass
import difflib
from enum import Enum
from pathlib import Path

from yaraast.ast.base import YaraFile
from yaraast.codegen.generator import CodeGenerator
from yaraast.parser.parser import Parser


class DiffType(Enum):
    """Type of difference."""

    ADD = "+"
    REMOVE = "-"
    MODIFY = "~"
    CONTEXT = " "


@dataclass
class DiffLine:
    """Represents a line in the diff."""

    type: DiffType
    line_num: int
    content: str
    old_content: str | None = None
    new_content: str | None = None


@dataclass
class DiffResult:
    """Result of a diff operation."""

    has_changes: bool
    lines: list[DiffLine]
    summary: dict[str, int]


class SimpleDiffer:
    """Simple differ for YARA files."""

    def __init__(
        self,
        parser: Parser | None = None,
        generator: CodeGenerator | None = None,
    ) -> None:
        """Initialize the differ."""
        self.parser = parser or Parser()
        self.generator = generator or CodeGenerator()

    def diff(self, content1: str, content2: str) -> DiffResult:
        """Diff two YARA file contents using LCS-based sequence matching."""
        lines1 = content1.splitlines()
        lines2 = content2.splitlines()

        diff_lines: list[DiffLine] = []
        added = 0
        removed = 0
        modified = 0

        matcher = difflib.SequenceMatcher(None, lines1, lines2)
        line_num = 0

        for tag, i1, i2, j1, j2 in matcher.get_opcodes():
            if tag == "equal":
                new_lines, line_num = _process_equal(lines1, i1, i2, line_num)
                diff_lines.extend(new_lines)
            elif tag == "replace":
                new_lines, line_num, a, r, m = _process_replace(
                    lines1, lines2, i1, i2, j1, j2, line_num
                )
                diff_lines.extend(new_lines)
                added += a
                removed += r
                modified += m
            elif tag == "insert":
                new_lines, line_num, a = _process_insert(lines2, j1, j2, line_num)
                diff_lines.extend(new_lines)
                added += a
            elif tag == "delete":
                new_lines, line_num, r = _process_delete(lines1, i1, i2, line_num)
                diff_lines.extend(new_lines)
                removed += r

        has_changes = added > 0 or removed > 0 or modified > 0

        return DiffResult(
            has_changes=has_changes,
            lines=diff_lines,
            summary={
                "added": added,
                "removed": removed,
                "modified": modified,
                "total_changes": added + removed + modified,
            },
        )


def _process_equal(
    lines1: list[str], i1: int, i2: int, line_num: int
) -> tuple[list[DiffLine], int]:
    """Process an 'equal' opcode block, returning context lines."""
    diff_lines: list[DiffLine] = []
    for idx in range(i2 - i1):
        line_num += 1
        diff_lines.append(
            DiffLine(
                type=DiffType.CONTEXT,
                line_num=line_num,
                content=f"  {lines1[i1 + idx]}",
                old_content=lines1[i1 + idx],
                new_content=lines1[i1 + idx],
            ),
        )
    return diff_lines, line_num


def _process_replace(
    lines1: list[str],
    lines2: list[str],
    i1: int,
    i2: int,
    j1: int,
    j2: int,
    line_num: int,
) -> tuple[list[DiffLine], int, int, int, int]:
    """Process a 'replace' opcode block, returning diff lines and counts."""
    diff_lines: list[DiffLine] = []
    added = 0
    removed = 0
    modified = 0
    old_chunk = lines1[i1:i2]
    new_chunk = lines2[j1:j2]
    for idx in range(max(len(old_chunk), len(new_chunk))):
        line_num += 1
        if idx < len(old_chunk) and idx < len(new_chunk):
            diff_lines.append(
                DiffLine(
                    type=DiffType.MODIFY,
                    line_num=line_num,
                    content=f"~ {new_chunk[idx]}",
                    old_content=old_chunk[idx],
                    new_content=new_chunk[idx],
                ),
            )
            modified += 1
        elif idx < len(new_chunk):
            diff_lines.append(
                DiffLine(
                    type=DiffType.ADD,
                    line_num=line_num,
                    content=f"+ {new_chunk[idx]}",
                    new_content=new_chunk[idx],
                ),
            )
            added += 1
        else:
            diff_lines.append(
                DiffLine(
                    type=DiffType.REMOVE,
                    line_num=line_num,
                    content=f"- {old_chunk[idx]}",
                    old_content=old_chunk[idx],
                ),
            )
            removed += 1
    return diff_lines, line_num, added, removed, modified


def _process_insert(
    lines2: list[str], j1: int, j2: int, line_num: int
) -> tuple[list[DiffLine], int, int]:
    """Process an 'insert' opcode block, returning added lines."""
    diff_lines: list[DiffLine] = []
    count = j2 - j1
    for idx in range(count):
        line_num += 1
        diff_lines.append(
            DiffLine(
                type=DiffType.ADD,
                line_num=line_num,
                content=f"+ {lines2[j1 + idx]}",
                new_content=lines2[j1 + idx],
            ),
        )
    return diff_lines, line_num, count


def _process_delete(
    lines1: list[str], i1: int, i2: int, line_num: int
) -> tuple[list[DiffLine], int, int]:
    """Process a 'delete' opcode block, returning removed lines."""
    diff_lines: list[DiffLine] = []
    count = i2 - i1
    for idx in range(count):
        line_num += 1
        diff_lines.append(
            DiffLine(
                type=DiffType.REMOVE,
                line_num=line_num,
                content=f"- {lines1[i1 + idx]}",
                old_content=lines1[i1 + idx],
            ),
        )
    return diff_lines, line_num, count


@dataclass
class ASTDiffResult:
    """Result of an AST diff operation."""

    has_changes: bool
    change_summary: dict[str, int]
    added_rules: list[str]
    removed_rules: list[str]
    modified_rules: list[str]
    logical_changes: list[str]
    structural_changes: list[str]
    style_only_changes: list[str]


class SimpleASTDiffer(SimpleDiffer):
    """Simplified AST differ for CLI use."""

    def __init__(
        self,
        parser: Parser | None = None,
        generator: CodeGenerator | None = None,
    ) -> None:
        super().__init__(parser=parser, generator=generator)

    def diff_files(self, file1: Path, file2: Path) -> ASTDiffResult:
        content1 = file1.read_text()
        content2 = file2.read_text()

        ast1 = self.parser.parse(content1)
        ast2 = self.parser.parse(content2)

        rules1 = {rule.name: rule for rule in ast1.rules}
        rules2 = {rule.name: rule for rule in ast2.rules}

        added_rules = sorted(set(rules2) - set(rules1))
        removed_rules = sorted(set(rules1) - set(rules2))

        modified_rules: list[str] = []
        for name in set(rules1) & set(rules2):
            if repr(rules1[name]) != repr(rules2[name]):
                modified_rules.append(name)

        logical_changes = []
        for name in added_rules:
            logical_changes.append(f"Rule added: {name}")
        for name in removed_rules:
            logical_changes.append(f"Rule removed: {name}")
        for name in modified_rules:
            logical_changes.append(f"Rule modified: {name}")

        change_summary = {
            "added_rules": len(added_rules),
            "removed_rules": len(removed_rules),
            "modified_rules": len(modified_rules),
        }

        has_changes = bool(added_rules or removed_rules or modified_rules)

        return ASTDiffResult(
            has_changes=has_changes,
            change_summary=change_summary,
            added_rules=added_rules,
            removed_rules=removed_rules,
            modified_rules=modified_rules,
            logical_changes=logical_changes,
            structural_changes=[],
            style_only_changes=[],
        )

    def diff_ast(self, ast1: YaraFile, ast2: YaraFile) -> DiffResult:
        """Diff two ASTs by comparing their generated code."""
        code1 = self.generator.generate(ast1)
        code2 = self.generator.generate(ast2)

        return self.diff(code1, code2)

    def get_changes(self, content1: str, content2: str) -> list[str]:
        """Get a list of changes between two contents."""
        result = self.diff(content1, content2)

        changes = []
        for line in result.lines:
            if line.type != DiffType.CONTEXT:
                changes.append(line.content)

        return changes

    def diff_directories(
        self,
        dir1: str | Path,
        dir2: str | Path,
    ) -> dict[str, DiffResult | ASTDiffResult]:
        """Diff all YARA files in two directories."""
        dir1 = Path(dir1)
        dir2 = Path(dir2)

        results: dict[str, DiffResult | ASTDiffResult] = {}

        # Get all .yar files in both directories
        files1 = {p.relative_to(dir1) for p in dir1.glob("**/*.yar")}
        files2 = {p.relative_to(dir2) for p in dir2.glob("**/*.yar")}

        # Files in both directories
        for file in files1 & files2:
            results[str(file)] = self.diff_files(dir1 / file, dir2 / file)

        # Files only in dir1 (removed)
        for file in files1 - files2:
            results[str(file)] = _diff_result_for_removed_file(dir1 / file)

        # Files only in dir2 (added)
        for file in files2 - files1:
            results[str(file)] = _diff_result_for_added_file(dir2 / file)

        return results


def _diff_result_for_removed_file(file_path: Path) -> DiffResult:
    """Create a DiffResult showing all lines of a file as removed."""
    with open(file_path) as f:
        content = f.read()
    lines = content.splitlines()
    diff_lines = [
        DiffLine(
            type=DiffType.REMOVE,
            line_num=i + 1,
            content=f"- {line}",
            old_content=line,
        )
        for i, line in enumerate(lines)
    ]
    return DiffResult(
        has_changes=True,
        lines=diff_lines,
        summary={"added": 0, "removed": len(lines), "modified": 0},
    )


def _diff_result_for_added_file(file_path: Path) -> DiffResult:
    """Create a DiffResult showing all lines of a file as added."""
    with open(file_path) as f:
        content = f.read()
    lines = content.splitlines()
    diff_lines = [
        DiffLine(
            type=DiffType.ADD,
            line_num=i + 1,
            content=f"+ {line}",
            new_content=line,
        )
        for i, line in enumerate(lines)
    ]
    return DiffResult(
        has_changes=True,
        lines=diff_lines,
        summary={"added": len(lines), "removed": 0, "modified": 0},
    )


def diff_lines(lines1: list[str], lines2: list[str]) -> list[DiffLine]:
    """Diff two lists of lines."""
    differ = SimpleDiffer()
    content1 = "\n".join(lines1)
    content2 = "\n".join(lines2)
    result = differ.diff(content1, content2)
    return result.lines


def diff_tokens(content1: str, content2: str) -> list[str]:
    """Diff tokens in two contents, accounting for frequency."""
    counts1 = Counter(content1.split())
    counts2 = Counter(content2.split())

    changes = []
    all_tokens = sorted(set(counts1) | set(counts2))
    for token in all_tokens:
        c1 = counts1.get(token, 0)
        c2 = counts2.get(token, 0)
        if c1 > c2:
            for _ in range(c1 - c2):
                changes.append(f"- {token}")
        elif c2 > c1:
            for _ in range(c2 - c1):
                changes.append(f"+ {token}")

    return changes


def diff_ast(ast1: YaraFile, ast2: YaraFile) -> DiffResult:
    """Diff two ASTs."""
    differ = SimpleASTDiffer()
    return differ.diff_ast(ast1, ast2)


def get_diff_summary(diff_result: DiffResult) -> dict[str, int]:
    """Get summary of diff result."""
    return diff_result.summary


def format_diff(diff_result: DiffResult) -> str:
    """Format diff result as string."""
    lines = []

    if not diff_result.has_changes:
        return "No changes"

    for line in diff_result.lines:
        if line.type != DiffType.CONTEXT:
            lines.append(line.content)

    return "\n".join(lines)


def print_diff(diff_result: DiffResult) -> None:
    """Print diff result to stdout."""
    print(format_diff(diff_result))



"""Registration helpers for the root CLI command tree."""

from __future__ import annotations

import click

from yaraast.cli.commands.analyze import analyze
from yaraast.cli.commands.bench_cmd import bench
from yaraast.cli.commands.diff_cmd import diff
from yaraast.cli.commands.fluent import fluent
from yaraast.cli.commands.fmt_cmd import fmt
from yaraast.cli.commands.format_cmd import format_yara, validate_syntax
from yaraast.cli.commands.libyara_cmd import libyara
from yaraast.cli.commands.metrics import metrics
from yaraast.cli.commands.optimize import optimize_cmd
from yaraast.cli.commands.parse_cmd import parse
from yaraast.cli.commands.performance import performance
from yaraast.cli.commands.performance_check import performance_check_cmd
from yaraast.cli.commands.roundtrip import roundtrip
from yaraast.cli.commands.semantic import semantic
from yaraast.cli.commands.serialize import serialize
from yaraast.cli.commands.validate import validate
from yaraast.cli.commands.workspace import workspace
from yaraast.cli.commands.yaral import yaral
from yaraast.cli.commands.yarax import yarax

CLI_COMMANDS = (
    parse,
    (format_yara, "format"),
    fmt,
    validate,
    validate_syntax,
    analyze,
    metrics,
    semantic,
    performance,
    performance_check_cmd,
    bench,
    diff,
    optimize_cmd,
    libyara,
    serialize,
    roundtrip,
    yaral,
    yarax,
    workspace,
    fluent,
)


def register_commands(cli: click.Group) -> None:
    """Register all top-level commands on the root Click group."""
    for command in CLI_COMMANDS:
        if isinstance(command, tuple):
            cli.add_command(command[0], name=command[1])
        else:
            cli.add_command(command)
    try:
        from yaraast.cli.commands.lsp import lsp

        cli.add_command(lsp)
    except ImportError:
        pass



"""Layout helpers for AdvancedCodeGenerator."""

from __future__ import annotations

from yaraast.codegen.formatting import BraceStyle, StringStyle
from yaraast.codegen.generator import CodeGenerator


def visit_yara_file(generator, node) -> str:
    imports = (
        sorted(node.imports, key=lambda item: item.module)
        if generator.config.sort_imports
        else node.imports
    )
    for imp in imports:
        generator.visit(imp)
    if imports:
        generator._write_blank_lines(generator.config.blank_lines_between_sections)

    for inc in node.includes:
        generator.visit(inc)
    if node.includes:
        generator._write_blank_lines(generator.config.blank_lines_between_sections)

    rules = node.rules
    if generator.config.sort_rules:
        rules = sorted(rules, key=lambda item: item.name)
    elif generator.config.sort_meta:

        def sort_key(rule):
            has_meta = bool(rule.meta and (bool(rule.meta)))
            return (not has_meta, rule.name)

        rules = sorted(rules, key=sort_key)

    for index, rule in enumerate(rules):
        if index > 0:
            generator._write_blank_lines(generator.config.blank_lines_between_rules)
        generator.visit(rule)
    return generator.buffer.getvalue()


def visit_rule(generator, node) -> str:
    if node.modifiers:
        generator._write(" ".join(str(m) for m in node.modifiers) + " ")
    generator._write(f"rule {node.name}")

    if node.tags:
        if generator.config.space_before_colon:
            generator._write(" ")
        generator._write(":")
        if generator.config.space_after_colon:
            generator._write(" ")
        tags_str = []
        for tag in node.tags:
            tags_str.append(
                tag if isinstance(tag, str) else tag.name if hasattr(tag, "name") else str(tag)
            )
        generator._write(" ".join(tags_str))

    if generator.config.brace_style == BraceStyle.SAME_LINE:
        generator._write(" {")
        generator._writeline()
    else:
        generator._writeline()
        generator._writeline("{")

    generator._indent()
    sections_written = 0
    for section in generator.config.section_order:
        if section == "meta" and node.meta:
            if sections_written > 0:
                generator._write_blank_lines(generator.config.blank_lines_between_sections)
            generator._write_meta_section(node.meta)
            sections_written += 1
        elif section == "strings" and node.strings:
            if sections_written > 0:
                generator._write_blank_lines(generator.config.blank_lines_between_sections)
            generator._write_strings_section(node.strings)
            sections_written += 1
        elif section == "condition":
            if sections_written > 0:
                generator._write_blank_lines(generator.config.blank_lines_between_sections)
            generator._write_condition_section(node.condition)
            sections_written += 1

    generator._dedent()
    generator._write("}")
    return generator.buffer.getvalue()


def write_strings_section(generator, strings) -> None:
    generator._writeline("strings:")
    generator._indent()
    if generator.config.sort_strings:
        strings = sorted(strings, key=lambda item: item.identifier)

    if generator.config.string_style in (StringStyle.ALIGNED, StringStyle.TABULAR):
        generator._collect_string_definitions(strings)
        write_aligned_strings(generator)
    else:
        for string_def in strings:
            generator.visit(string_def)
            generator._writeline()
    generator._dedent()


def write_aligned_strings(generator) -> None:
    if not generator._string_definitions:
        return
    max_id_len = max(len(identifier) for identifier, _, _ in generator._string_definitions)
    max_val_len = max(len(value) for _, value, _ in generator._string_definitions)
    for identifier, value, modifiers in generator._string_definitions:
        generator._write(generator._get_indent())
        if generator.config.string_style == StringStyle.TABULAR:
            generator._write(identifier.ljust(max_id_len))
            generator._write(" = ")
            generator._write(value.ljust(max_val_len))
        else:
            generator._write(f"{identifier} = {value}")
        if modifiers:
            generator._write("  " if generator.config.align_string_modifiers else " ")
            generator._write(" ".join(modifiers))
        generator._writeline()


def write_condition_section(generator, condition) -> None:
    generator._writeline("condition:")
    generator._indent()
    condition_str = generate_condition_string(condition)
    if len(condition_str) > generator.config.max_line_length:
        generator._writeline(condition_str)
    else:
        generator._writeline(condition_str)
    generator._dedent()


def generate_condition_string(expr) -> str:
    temp_gen = CodeGenerator()
    return temp_gen.visit(expr)



"""Outcome section validation mixin for YARA-L."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from yaraast.yaral.ast_nodes import (
        AggregationFunction,
        OutcomeAssignment,
        OutcomeExpression,
        OutcomeSection,
    )


class OutcomeValidationMixin:
    """Validate outcome assignments and aggregations."""

    def _validate_outcome_section(self, node: OutcomeSection) -> None:
        """Validate outcome section."""
        assignments = node.assignments if hasattr(node, "assignments") else []

        if hasattr(node, "variables"):
            for var_name in node.variables:
                self._register_outcome_variable(var_name)

        for assignment in assignments:
            self._register_outcome_variable(assignment.variable)

    def _register_outcome_variable(self, var_name: str) -> None:
        if var_name in self.defined_outcome_vars:
            self._add_error(
                "outcome",
                f"Duplicate outcome variable: {var_name}",
                "Use unique outcome variable names",
            )
        self.defined_outcome_vars.add(var_name)

        reserved = ["risk_score", "severity", "confidence", "priority"]
        if var_name not in reserved and not var_name.startswith("$"):
            self._add_error(
                "outcome",
                f"Outcome variable '{var_name}' must start with $ or use a reserved name",
                f"Reserved names: {', '.join(reserved)}",
            )

    def visit_yaral_outcome_section(self, node: OutcomeSection) -> None:
        for assignment in node.assignments:
            self.visit(assignment)

    def visit_yaral_outcome_assignment(self, node: OutcomeAssignment) -> None:
        self.defined_outcome_vars.add(node.variable)
        if hasattr(node.expression, "accept"):
            self.visit(node.expression)

    def visit_yaral_outcome_expression(self, node: OutcomeExpression) -> None:
        return

    def visit_yaral_aggregation_function(self, node: AggregationFunction) -> None:
        if node.function not in self.VALID_AGGREGATIONS:
            self._add_warning(
                "outcome",
                f"Unknown aggregation function: {node.function}",
                f"Valid functions: {', '.join(self.VALID_AGGREGATIONS)}",
            )
